/*
 *   Generic filesystem routines and wrappers
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <string.h>
#else
#include <strings.h>
#include <unistd.h>
#endif



/*
static FILE *fpx;
#define LOGULE(x,y) {fpx=fopen("qslog","a");if(!fpx) return; fprintf(fpx,x,y); fclose(fpx);}
*/

#include "ithelib.h"

// Defines

#define MAXIFILES 8192

// Variables

static struct IFILEENTRY
	{
	char *name;
	long org;
	long len;
	char *tfn;
	} ifiletab[MAXIFILES];
static int ifilenum=0;

// Functions

static int BSCMP(const void *key, const void *elem);
static int QSCMP(const void *elem1, const void *elem2);

// Code

IFILE *iopen(char *filename)
{
FILE *fp;
IFILE *ifp;
struct IFILEENTRY *ife;

// Try a physical file.  If it doesn't exist, try the VFS table

fp = fopen(filename,"rb");
if(!fp)
	{
	// Search the VFS table, and fill out the entries if it worked
	ife = bsearch(filename,ifiletab,ifilenum,sizeof(struct IFILEENTRY),BSCMP);
	if(ife)
		{
		ifp = (IFILE *)M_get(1,sizeof(IFILE));
		ifp->length = ife->len;
		ifp->origin = ife->org;
		fp = fopen(ife->tfn,"rb");
		if(!fp)
			ithe_panic("iopen: fopen failed on file",ife->tfn);
		fseek(fp,ifp->origin,SEEK_SET);
		ifp->fp = fp;
		ifp->truename = M_get(1,strlen(ife->tfn)+1);
		strcpy(ifp->truename,ife->tfn);
		return ifp;
		}

	// Bugger.
	ithe_panic("iopen: Could not open file",filename);
//	return NULL;
    }

// OK, we got a result with a physical file, load it and go
ifp = (IFILE *)M_get(1,sizeof(IFILE));
ifp->length = filelength(fileno(fp));
ifp->origin = 0;
ifp->fp = fp;

ifp->truename = M_get(1,strlen(filename)+1);
strcpy(ifp->truename,filename);

return ifp;
}

IFILE *iopen_write(char *filename)
{
FILE *fp;
IFILE *ifp;

// only physical files are supported for writing

fp = fopen(filename,"wb+");
if(!fp)
	ithe_panic("iopen_write: cannot create file:",filename);

// OK, we got a result with a physical file, start it and go
ifp = (IFILE *)M_get(1,sizeof(IFILE));
ifp->length = 0;
ifp->origin = 0;
ifp->fp = fp;

ifp->truename = M_get(1,strlen(filename)+1);
strcpy(ifp->truename,filename);

return ifp;
}

int iexist(char *filename)
{
if(!access(filename,F_OK))
	return 1;
if(bsearch(filename,ifiletab,ifilenum,sizeof(struct IFILEENTRY),BSCMP))
	return 1; // It is here
return 0;
}


int ifile_attach(char *filename, char *truename, long length, long origin)
{
if(bsearch(filename,ifiletab,ifilenum,sizeof(struct IFILEENTRY),BSCMP))
	return 0; // Already there, don't bother

if(ifilenum > MAXIFILES)
	ithe_panic("ifile_attach: Too many files",filename);

ifiletab[ifilenum].name = filename;
ifiletab[ifilenum].org = origin;
ifiletab[ifilenum].tfn = truename;
ifiletab[ifilenum++].len = length;

qsort(ifiletab,ifilenum,sizeof(struct IFILEENTRY),QSCMP);

return 1;
}

void iclose(IFILE *ifp)
{
if(!ifp)
	return;
fclose(ifp->fp);
M_free(ifp->truename);
M_free(ifp);
}

void iseek(IFILE *ifp, int offset, int whence)
{
if(!ifp)
	return;

switch(whence)
	{
	case SEEK_SET:
	fseek(ifp->fp,offset+ifp->origin,SEEK_SET);
	break;

	case SEEK_END:
	fseek(ifp->fp,ifp->origin+ifp->length+offset,SEEK_SET);
	break;

	default:
	fseek(ifp->fp,offset,SEEK_CUR);
	break;
	}
}

unsigned char igetc(IFILE *ifp)
{
return fgetc(ifp->fp);
}

unsigned short igetsh_reverse(IFILE *ifp)
{
return ((igetc(ifp)<<8) + igetc(ifp));
}

unsigned short igetsh_native(IFILE *ifp)
{
unsigned short s;
iread((void *)&s,2,ifp);
return s;
}

unsigned long igetl_reverse(IFILE *ifp)
{
return ((igetsh_reverse(ifp) << 16) + igetsh_reverse(ifp));
}

unsigned long igetl_native(IFILE *ifp)
{
return getw(ifp->fp);
}

uqword igetll_reverse(IFILE *ifp)
{
return (((uqword)igetl_reverse(ifp) << 32) + igetl_reverse(ifp));
}

uqword igetll_native(IFILE *ifp)
{
uqword q;
iread((void *)&q,8,ifp);
return q;
}

int iread(unsigned char *buf, int l, IFILE *ifp)
{
return fread(buf,1,l,ifp->fp);
}

void iputc(unsigned char c, IFILE *ifp)
{
fputc(c,ifp->fp);
}

void iputsh_reverse(unsigned short s, IFILE *ifp)
{
iputc((s>>8)&0xff,ifp);
iputc(s&0xff,ifp);
}

void iputsh_native(unsigned short s, IFILE *ifp)
{
iwrite((void *)&s,2,ifp);
}

void iputl_reverse(unsigned long l, IFILE *ifp)
{
iputsh_reverse((l>>16)&0xffff,ifp);
iputsh_reverse(l&0xffff,ifp);
}

void iputl_native(unsigned long l, IFILE *ifp)
{
putw(l,ifp->fp);
}

void iputll_reverse(uqword q, IFILE *ifp)
{
iputl_reverse((q>>32)&0xffffffff,ifp);
iputl_reverse(q&0xffffffff,ifp);
}

void iputll_native(uqword q, IFILE *ifp)
{
iwrite((void *)&q,8,ifp);
}

int iwrite(unsigned char *buf, int l, IFILE *ifp)
{
return fwrite(buf,1,l,ifp->fp);
}

unsigned int itell(IFILE *ifp)
{
return ftell(ifp->fp)-ifp->origin;
}

unsigned int ifilelength(IFILE *ifp)
{
return ifp->length;
}

int ieof(IFILE *ifp)
{
if(ftell(ifp->fp) >= ifp->origin+ifp->length)
	return 1;
return 0;
}

static int BSCMP(const void *key, const void *elem)
{
return stricmp((char *)key,((struct IFILEENTRY *)elem)->name);
}

static int QSCMP(const void *elem1, const void *elem2)
{
/*
LOGULE("elem1=%x\n",elem1);
LOGULE("elem1->name=%x\n",((struct IFILEENTRY *)elem1)->name);
LOGULE("elem1->name=%s\n",((struct IFILEENTRY *)elem1)->name);
LOGULE("elem2=%x\n",elem2);
LOGULE("elem2->name=%x\n",((struct IFILEENTRY *)elem2)->name);
LOGULE("elem2->name=%s\n",((struct IFILEENTRY *)elem2)->name);
*/
return stricmp(((struct IFILEENTRY *)elem1)->name,((struct IFILEENTRY *)elem2)->name);
}



#ifdef __BEOS__		// BEOS is not well
int getw(FILE *fp)
{
int w;
fread(&w,1,4,fp);
return w;
}

void putw(int w, FILE *fp)
{
fwrite(&w,1,4,fp);
}
#endif
