/*
 *      Optimised A-Star algorithm by Patrick Hogan (Patter@iname.com)
 *
 *
 *      Code heavily modified for use with IRE.  The core is the same.
 */

#include "ithelib.h"
#include "media.h"
#include "core.hpp"
#include "gamedata.h"
#include "console.h"
#include "object.hpp"
#include <math.h>
#include <string.h>

#define MAPSIZE 64

//#define heuristic(x1,y1,x2,y2) 10*fast_sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1))

#define heuristic(x1,y1,x2,y2) 10*sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1))

static BITMAP *imap;
static char imap_on=0;
extern "C" char show_imap;
extern "C" int imapx,imapy;

extern void gen_largemap();
extern void CentreMap(OBJECT *centre);

/*
Tile class, structure for individual squares on grid map.
*/
class AST_TILE {
 public:
  AST_TILE *parent;
  int cost, h;
  int search;
  int wall;
  int x, y;
  AST_TILE() { h = 0; cost = 0; wall = 1; search = 0; parent = NULL; }
};

/*
Node class for ordered stack.
*/
class AST_NODE {
 public:
  AST_TILE *tile;
  AST_NODE *next;
};

/*
Class wrapper for astar algorithm.
*/
class ASTAR {
 private:
  AST_NODE *first;
  int search;
 public:
  AST_TILE pfmap[MAPSIZE][MAPSIZE];
  ASTAR();
  void add_node(AST_TILE *tile);
  AST_TILE *lift_node();
  void clear_nodes();
  AST_TILE *generate8(int x1, int y1, int x2, int y2);
  AST_TILE *generate4(int x1, int y1, int x2, int y2);
};


extern int AlwaysSolid(int x,int y);
extern int isSolid(int x,int y);
extern int IsTileSolid(int x,int y);
extern void OB_SetDir(OBJECT *objsel,int dir);
extern OBJECT *GetRawSolidObject(int x,int y);
extern void CallVMnum(int v);
static void do_dir(OBJECT *o,int nx,int ny);

/*
Constructor
*/

ASTAR::ASTAR()
{
  first = NULL;
  search = 0;
//  memcpy(tilemap,&pfmap,MAPSIZE*MAPSIZE*sizeof(AST_TILE));
}

/*
Node adding function for an ordered stack.
*/
void ASTAR::add_node(AST_TILE *tile) {
  AST_NODE **rp = &first;
  AST_NODE *p = first;
  AST_NODE *insert = new AST_NODE;
  insert->tile = tile;
  while (p && p->tile->h < tile->h) {
    rp = &p->next;
    p = p->next;
  }
  *rp = insert;
  insert->next = p;
}

/*
Retrieve the next node from the stack.
*/
AST_TILE *ASTAR::lift_node() {
  AST_NODE *p = first;
  AST_TILE *t = first->tile;
  first = first->next;
  delete(p);
  return t;
}

/*
Clear all nodes on the stack.
*/
void ASTAR::clear_nodes() {
  AST_NODE *p = first;
  AST_NODE *save = NULL;
  while (p != NULL) {
    save = p->next;
    delete(p);
    p = save;
  }
  first = NULL;
}

/*
The core search function. The real down-to-earth-nitty-gritty of the search
engine. Implements a basic optimised A* algorithm. If the search is
successful it returns a pointer to the last tile which can be traced back
to the start by the moving (tile = tile->parent) until tile == NULL;
Search returns NULL if search failed.
*/

AST_TILE *ASTAR::generate8(int x1, int y1, int x2, int y2) {
  int x[8] = { 0, 1, 1, 1, 0,-1,-1,-1};
  int y[8] = {-1,-1, 0, 1, 1, 1, 0,-1};
  int cost[8] = {10,14,10,14,10,14,10,14};
  int i,h;
  AST_TILE *insert, *expand;

  search++;
  clear_nodes();
  insert = &(pfmap[x1][y1]);
  insert->parent = NULL;
  insert->search = search;
  add_node(insert);
  while (1) {
    /* if the stack is empty then we've exhausted all possible nodes. */
    if (first == NULL)
       return NULL;

    expand = lift_node();

    if (expand->x == x2 && expand->y == y2)
      return expand; /* Yes! Found it! Return pointer to last tile. */

    for (i=0; i<8; i++) {
      insert = &(pfmap[expand->x+x[i]][expand->y+y[i]]);
      if (insert->wall == 0) {
        h = (int)(heuristic(insert->x,insert->y,x2,y2));
 	if (insert->search != search || insert->cost > expand->cost+cost[i]+h) {
          insert->search = search;
 	  insert->parent = expand;
 	  insert->h = cost[i] + h;
 	  insert->cost = expand->cost + insert->h;
 	  add_node(insert);
 	}
      }
    }
  }
}


AST_TILE *ASTAR::generate4(int x1, int y1, int x2, int y2) {
  int x[4] = { 0, 1,  0,-1};
  int y[4] = {-1, 0,  1, 0};
  int cost[8] = {10,10,10,10,10,10,10,10};
	int i,h;
  AST_TILE *insert, *expand;

  search++;
  clear_nodes();
  insert = &(pfmap[x1][y1]);
  insert->parent = NULL;
  insert->search = search;
  add_node(insert);
  while (1) {
    /* if the stack is empty then we've exhausted all possible nodes. */
    if (first == NULL)
       return NULL;

    expand = lift_node();

    if (expand->x == x2 && expand->y == y2)
      return expand; /* Yes! Found it! Return pointer to last tile. */

    for (i=0; i<4; i++) {
      insert = &(pfmap[expand->x+x[i]][expand->y+y[i]]);
      if (insert->wall == 0) {
        h = (int)(heuristic(insert->x,insert->y,x2,y2));
 	if (insert->search != search || insert->cost > expand->cost+cost[i]+h) {
          insert->search = search;
 	  insert->parent = expand;
 	  insert->h = cost[i] + h;
 	  insert->cost = expand->cost + insert->h;
 	  add_node(insert);
 	}
      }
    }
  }
}

/* END */

/*
 *      A-Star interface to the game engine
 */


int FindPath(OBJECT *start, OBJECT *end, int diagonal)
{
int vsx,vsy,oldx,oldy,tx,ty,cx,cy,xctr,yctr;
int blockx,blocky,blockw,blockh;
OBJECT *temp,*temp2;
ASTAR astar;
int xxx=0;
AST_TILE *p;

if(!imap_on)
    {
    imap = create_bitmap(MAPSIZE,MAPSIZE);
    imap_on = 1;
    }

// If the two objects are on top of each other, return code 0 (found it)

if(start->x == end->x && start->y == end->y)
    return 0;

// Centre the window around the START object

vsx = start->x-(MAPSIZE>>1);
if(vsx<0)
    vsx=0;
vsy = start->y-(MAPSIZE>>1);
if(vsy<0)
    vsy=0;

// Detect an out-of-bounds target and return abort (code 0)

if(end->x<=vsx || end->y<=vsy || end->x>=(vsx+MAPSIZE-1) || end->y>=(vsy+MAPSIZE-1))
    {
//    C_printf("Bounds Error\n");
    return 0;
    }

// Draw a border around the edges of the map
// First the horzontal

for(cx=0;cx<MAPSIZE;cx++)
    {
    astar.pfmap[cx][0].x = cx;
    astar.pfmap[cx][0].y = 0;
    astar.pfmap[cx][0].wall = 1;
    astar.pfmap[cx][MAPSIZE-1].x = cx;
    astar.pfmap[cx][MAPSIZE-1].y = MAPSIZE-1;
    astar.pfmap[cx][MAPSIZE-1].wall = 1;
    }

// Then the vertical

for(cy=0;cy<MAPSIZE;cy++)
    {
    astar.pfmap[0][cy].x = 0;
    astar.pfmap[0][cy].y = cy;
    astar.pfmap[0][cy].wall = 1;
    astar.pfmap[MAPSIZE-1][cy].x = MAPSIZE-1;
    astar.pfmap[MAPSIZE-1][cy].y = cy;
    astar.pfmap[MAPSIZE-1][cy].wall = 1;
    }

/*
for(cx=0;cx<MAPSIZE;cx++)
    for(cy=0;cy<MAPSIZE;cy++)
        {
        pfmap[cx][cy].x = cx;
        pfmap[cx][cy].y = cy;
        pfmap[cx][cy].wall = 1;
        }
*/

// Now, fill in the actual map with details of the environment

for(cx=1;cx<MAPSIZE-1;cx++)
    {
    for(cy=1;cy<MAPSIZE-1;cy++)
        {
        astar.pfmap[cx][cy].x = cx;
        astar.pfmap[cx][cy].y = cy;
        astar.pfmap[cx][cy].wall = 0;

        tx = cx+vsx;
        ty = cy+vsy;

        // Get the solid object at this point
        temp = GetRawSolidObject(tx,ty);

        // If there is a solid map tile, mark as impassible
        if(IsTileSolid(tx,ty))
             {
             astar.pfmap[cx][cy].wall = 1;
             temp = NULL;           // Don't waste time checking the objects
             }

        // Check hte objects.
        if(temp)
          if(!temp->flags.willopen)
             {
             if(temp->flags.large)      // It's large, do lots of stuff
                 {
                 // First, generate the appropriate object shape
                 if(temp->curdir>CHAR_D)
                     {
                     blockx=temp->hblock[BLK_X];
                     blocky=temp->hblock[BLK_Y];
                     blockw=temp->hblock[BLK_W];
                     blockh=temp->hblock[BLK_H];
                     }
                 else
                     {
                     blockx=temp->vblock[BLK_X];
                     blocky=temp->vblock[BLK_Y];
                     blockw=temp->vblock[BLK_W];
                     blockh=temp->vblock[BLK_H];
                     }

		if(cx+blockw+blockx>MAPSIZE)
			blockw-=((cx+blockw+blockx)-MAPSIZE);
		if(cy+blockh+blocky>MAPSIZE)
			blockh-=((cy+blockh+blocky)-MAPSIZE);

                 // Fill in the map where the solid parts of the shape are

                 for(xctr=0;xctr<blockw;xctr++)
                     for(yctr=0;yctr<blockh;yctr++)
                         astar.pfmap[cx+xctr+blockx][cy+yctr+blocky].wall=1;
                 }
             else
                 astar.pfmap[cx][cy].wall = 1;    // It's just a small one
             }
        }
    }

// Punch the dest and source out of the solidity map

astar.pfmap[end->x-vsx][end->y-vsy].wall=0;
astar.pfmap[start->x-vsx][start->y-vsy].wall=0;

// For debugging, draw the cute little map

if(show_imap)
    {
    for(cx=0;cx<MAPSIZE;cx++)
        for(cy=0;cy<MAPSIZE;cy++)
            putpixel(imap,cx,cy,astar.pfmap[cx][cy].wall?0:makecol(255,255,255));

    putpixel(imap,start->x-vsx,start->y-vsy,makecol(0,0,255));
    putpixel(imap,end->x-vsx,end->y-vsy,makecol(255,0,0));
    draw_sprite(swapscreen,imap,imapx,imapy);
    }

// request the path

if(diagonal)
    p = astar.generate8(start->x-vsx,start->y-vsy,end->x-vsx,end->y-vsy);
else
    p = astar.generate4(start->x-vsx,start->y-vsy,end->x-vsx,end->y-vsy);

// If it can't be done, return failure (code 2)

if(!p)
	{
//	ilog_printf("Failed %s->%s\n",start->name,end->name);
	return 2;
	}

if(p->parent)
    {
    for(;p->parent->parent;p=p->parent);        // Wind to the last useful bit
    do_dir(start,p->x+vsx,p->y+vsy);            // Make the object face it
    start->user->dx=(p->x+vsx)-start->x;        // Get delta x and y
    start->user->dy=(p->y+vsy)-start->y;
    // Try to move, if we fail, call a script to signal failure and
    // decide what to do about it with user-code.
    // This can open a door or whatever.

    if(!MoveObject(start,p->x+vsx,p->y+vsy))
        {

        // Ok, if it's a semi-solid object, we need to move the largemap
        // window to cover the moving object, not the player

        CentreMap(start);
        gen_largemap();

/*
        tempp = player;
        player = start;
        CentreMap(start);
        gen_largemap();
        player=tempp;
*/

        // Now let's try again

        if(!MoveObject(start,p->x+vsx,p->y+vsy))
            {
            temp = current_object;
            temp2 = victim;
            oldx = new_x;
            oldy = new_y;

            current_object = start;
            victim = end;
            new_x = p->x+vsx;
            new_y = p->y+vsy;
//ilog_printf("%s calling trackstop at %d,%d\n",current_object->name,new_x,new_y);
            CallVMnum(trackstop);

            current_object = temp;
            victim = temp2;
            new_x = oldx;
            new_y = oldy;
            }

        // Now move the window back over the player again

        CentreMap(player);
        gen_largemap();
        }
    }
else
    {
    Bug("pathfinder: No parent (please report this bug)\n");
    }

return 1;       // Return code 1 (moved OK, but not yet found the object)
}

/*
 *      Make the object face the proper direction
 */

void do_dir(OBJECT *o,int nx,int ny)
{
if(o->x<nx)
    OB_SetDir(o,CHAR_R);
if(o->x>nx)
    OB_SetDir(o,CHAR_L);
if(o->y<ny)
    OB_SetDir(o,CHAR_D);
if(o->y>ny)
    OB_SetDir(o,CHAR_U);
}

/*
 *      can_route - return the number of steps in the route, or -1
 */

int CanRoute(OBJECT *start, OBJECT *end, int diagonal)
{
int vsx,vsy,oldx,oldy,tx,ty,xctr,yctr,cx,cy;
int blockx,blocky,blockw,blockh;
OBJECT *temp,*temp2;
ASTAR astar;
AST_TILE *p;

// If the two objects are on top of each other, return 0 steps

if(start->x == end->x && start->y == end->y)
    return 0;

// Centre the window around the START object

vsx = start->x-(MAPSIZE>>1);
if(vsx<0)
    vsx=0;
vsy = start->y-(MAPSIZE>>1);
if(vsy<0)
    vsy=0;

// Detect an out-of-bounds target and return -1 steps

if(end->x<=vsx || end->y<=vsy || end->x>=(vsx+MAPSIZE-1) || end->y>=(vsy+MAPSIZE-1))
    return -1;

// Draw a border around the edges of the map
// First the horzontal

for(cx=0;cx<MAPSIZE;cx++)
    {
    astar.pfmap[cx][0].x = cx;
    astar.pfmap[cx][0].y = 0;
    astar.pfmap[cx][0].wall = 1;
    astar.pfmap[cx][MAPSIZE-1].x = cx;
    astar.pfmap[cx][MAPSIZE-1].y = MAPSIZE-1;
    astar.pfmap[cx][MAPSIZE-1].wall = 1;
    }

// Then the vertical

for(cy=0;cy<MAPSIZE;cy++)
    {
    astar.pfmap[0][cy].x = 0;
    astar.pfmap[0][cy].y = cy;
    astar.pfmap[0][cy].wall = 1;
    astar.pfmap[MAPSIZE-1][cy].x = MAPSIZE-1;
    astar.pfmap[MAPSIZE-1][cy].y = cy;
    astar.pfmap[MAPSIZE-1][cy].wall = 1;
    }

// Now, fill in the actual map with details of the environment

for(cx=1;cx<MAPSIZE-1;cx++)
    {
    for(cy=1;cy<MAPSIZE-1;cy++)
        {
        astar.pfmap[cx][cy].x = cx;
        astar.pfmap[cx][cy].y = cy;
        astar.pfmap[cx][cy].wall = 0;

        tx = cx+vsx;
        ty = cy+vsy;

        // Get the solid object at this point
        temp = GetRawSolidObject(tx,ty);

        // If there is a solid map tile, mark as impassible
        if(IsTileSolid(tx,ty))
             {
             astar.pfmap[cx][cy].wall = 1;
             temp = NULL;           // Don't waste time checking the objects
             }

        // Check the object.  If someone is in the way, assume they'll move
        if(temp)
          if(!temp->flags.willopen && !temp->flags.person)
             {
             if(temp->flags.large)      // It's large, do lots of stuff
                 {
                 // First, generate the appropriate object shape
                 if(temp->curdir>CHAR_D)
                     {
                     blockx=temp->hblock[BLK_X];
                     blocky=temp->hblock[BLK_Y];
                     blockw=temp->hblock[BLK_W];
                     blockh=temp->hblock[BLK_H];
                     }
                 else
                     {
                     blockx=temp->vblock[BLK_X];
                     blocky=temp->vblock[BLK_Y];
                     blockw=temp->vblock[BLK_W];
                     blockh=temp->vblock[BLK_H];
                     }

                 // Fill in the map where the solid parts of the shape are

                 for(xctr=0;xctr<blockw;xctr++)
                     for(yctr=0;yctr<blockh;yctr++)
                         astar.pfmap[cx+xctr+blockx][cy+yctr+blocky].wall=1;
                 }
             else
                 astar.pfmap[cx][cy].wall = 1;    // It's just a small one
             }
        }
    }

// Punch the dest and source out of the solidity map

astar.pfmap[end->x-vsx][end->y-vsy].wall=0;
astar.pfmap[start->x-vsx][start->y-vsy].wall=0;

// request the path

if(diagonal)
    p = astar.generate8(start->x-vsx,start->y-vsy,end->x-vsx,end->y-vsy);
else
    p = astar.generate4(start->x-vsx,start->y-vsy,end->x-vsx,end->y-vsy);

// If it can't be done, return failure

if(!p)
    return -1;

if(!p->parent)
    return -1;

cx=0;
for(;p->parent->parent;p=p->parent) cx++; // count the steps in the path

return cx;      // and return them
}
