/*
 *      Project - Sprite and map projection code
 */

#include <string.h>

#include "core.hpp"
#include "ithelib.h"
#include "loadfile.h"
#include "gamedata.h"
#include "darken.h"
#include "map.hpp"
#include "object.hpp"
#include "console.h"
#include "media.h"

// Defines
#define MAX_P_OVERLAY 255       // Maximum number of overlayed objects
                                // Per screen.  Somewhat unlikely..

struct LTAB
    {
    int x,y;
    int light;
    };

// Variables

static char Animate=0;
static unsigned char tip[65535];        // Tile Animation Pointer
static BITMAP **tsb;                     // Tile Scroll Buffers
static BITMAP *warning;
static BITMAP *darkmap;
static BITMAP *tsbscreen;
static int darklevel=128;
static COLOR_MAP *oldcmap;

void (*animate[4])(OBJECT *temp);
static unsigned char *rndtab[256];

    // Lookup table for the possible light paths.
    // Each node has three adjacent squares which are lit by the engine.
    // The dark routine lights up (or darkens) each square, and marks it
    // as being used, so that it cannot be brightened twice.
    // This is necessary because there will be overlap between some nodes.
    // That's necessary due to this algorithm...

    static int node[8][4][2]={
                        { {-1,-1},{-2,-2},{-2,-1},{-1,-2}  }, // top-left
                        { {+0,-1},{+0,-2},{-1,-2},{+1,-2}  }, // top
                        { {+1,-1},{+2,-2},{+1,-2},{+2,-1}  }, // top-right
                        { {+1,+0},{+2,+0},{+2,-1},{+2,+1}  }, // right
                        { {+1,+1},{+2,+2},{+2,+1},{+1,+2}  }, // bottom-right
                        { {+0,+1},{+0,+2},{-1,+2},{+1,+2}  }, // bottom
                        { {-1,+1},{-2,+2},{-2,+1},{+1,+2}  }, // bottom-left
                        { {-1,+0},{-2,+0},{-2,-1},{-2,+1}  }  // left
                      };

static BITMAP *lightmap;       // The shape of the lightmask we'll use
static BITMAP *lm[8][8];         // The lightmask split into tiles.
static BITMAP *lmcore;
static int lmw,lmh;            // Width and height of the lightmask in tiles.
static int lmw2,lmh2;          // lmw and lmh respectively, but divided by two
static int basex,basey;
static char vismap[LMSIZE][LMSIZE]; // visibility state
static char blmap[LMSIZE][LMSIZE];  // blocks light
int gamewinsize;

static OBJECT *roofsprites[MAX_P_OVERLAY]; // Objects which are on the roof
static int roofspr;
static int vis_black;

// Functions

void Init_Projector(int w, int h);
void Project_Map(int x, int y);
void Project_Sprites(int x, int y);
void Project_Roof(int x, int y);
void Hide_Unseen(int x, int y);
void SetDarkness(int d);

static void anim_00(OBJECT *temp);
static void anim_01(OBJECT *temp);
static void anim_10(OBJECT *temp);
static void anim_11(OBJECT *temp);
static void anim_random(OBJECT *temp);

static void init_lightmap();
static void proj_light(int x, int y);
void render_lighting(int x, int y);
static void vis_floodfill(int x, int y);

// Code

/*
 *      Init_Projector() - Set up the jump table for the animator
 *                         Create projection window
 */

void Init_Projector(int w, int h)
{
int ctr,ctr2;
char filename[1024];

vis_black=makecol(0,0,0);

animate[0] = anim_00;
animate[1] = anim_01;
animate[2] = anim_10;
animate[3] = anim_11;

tsbscreen = create_bitmap(64,64);
if(!tsbscreen)
    ithe_panic("Could not allocate tile scroll buffer",NULL);

gamewin = create_bitmap(w,h);
if(!gamewin)
    ithe_panic("Could not allocate projection window",NULL);
gamewinsize = w*h;

roofwin = create_bitmap(w,h);
if(!roofwin)
    ithe_panic("Could not allocate projection window",NULL);

tsb = (BITMAP **)M_get(TItot,sizeof(BITMAP *)); // Tile Scroll Buffer

if(!loadfile("warning.cel",filename))
	ithe_panic("Could not find Warning image (warning.cel)",NULL);
warning=iload_bitmap(filename);
if(!warning)
	ithe_panic("Could not load Warning image",filename);

// Build a random number table for the random animation code

for(ctr=0;ctr<256;ctr++)
    {
    rndtab[ctr]=(unsigned char *)M_get(1,512);
    if(ctr>0)
        for(ctr2=0;ctr2<512;ctr2++)
            {
            rndtab[ctr][ctr2]=rand()%ctr;
            // Prevent consecutive numbers appearing
            if(ctr2>0 && ctr>1)
                while(rndtab[ctr][ctr2]==rndtab[ctr][ctr2-1])
                    rndtab[ctr][ctr2]=rand()%ctr;
            }
    }

// Build the tile scroll buffer

for(ctr=0;ctr<TItot;ctr++)
	if(!tsb[ctr])
		{
		tsb[ctr]=create_bitmap(32,32);
		if(!tsb[ctr])
			{
			TIlist[ctr].sdx=0;
			TIlist[ctr].sdy=0;
			Bug("Could not create TSB for tile %d\n",ctr);
				return;
			}
		}


darkmap = create_bitmap_ex(8,VSW32,VSH32);
if(!darkmap)
	ithe_panic("Create DMP bitmap failed",NULL);

darken_init(ire_bpp);

init_lightmap();
SetDarkness(darklevel); // Resync darkness level
}

/*
 *      anim_00() - no ping-pong, no loop
 */

void anim_00(OBJECT *temp)
{
temp->sptr+=temp->sdir;         // Increment the animation frame
if(temp->sptr>=temp->slen)      // If the animation pointer > # of frames
	{
	temp->sptr=temp->slen-1;   // Set it to the last frame
	temp->sdir=0;              // Stop playing

        // If we're going to jump to another sequence, restart it
        if(temp->form->flags&8)
            {
            temp->sdir=1;
            temp->sptr=0;
            temp->slen = temp->form->jumpto->frames;
            temp->form = temp->form->jumpto;
            }
	}
}

/*
 *      anim_01() - ping-pong, no loop
 */

void anim_01(OBJECT *temp)
{
temp->sptr+=temp->sdir;         // Increment the animation frame
if(temp->sptr>=temp->slen)	// if we reach upper limit
	{
	temp->sdir=-1;          // Play backwards
	temp->sptr-=2;		// This prevents 0123321001233210 etc
	}									// (should be 012321012321 etc)
else
	if(temp->sptr<0)	// if we reach -1, we've finished the backrun
	{
	temp->sdir=0;		// shut down
	temp->sptr=0;		// NEVER have a cyclic loop with 1 frame
	                        // or this will blow up on you.

        // If we're going to jump to another sequence, restart it
        if(temp->form->flags&8)
            {
            temp->sdir=1;
            temp->sptr=0;
            temp->slen = temp->form->jumpto->frames;
            temp->form = temp->form->jumpto;
            }
	}

}

/*
 *      anim_10() - no ping-pong, loop
 */

void anim_10(OBJECT *temp)
{
temp->sptr+=temp->sdir;         // Increment the animation frame
if(temp->sptr>=temp->slen)      // If we reach the end,
	temp->sptr=0;           // Go back to the start
}

/*
 *      anim_11() - ping-pong, and loop
 */

void anim_11(OBJECT *temp)
{
temp->sptr+=temp->sdir;         // Increment the animation frame
if(temp->sptr>=temp->slen)	// if we reach upper limit
	{
	temp->sdir=-1;          // Play backwards
	temp->sptr-=2;		// prevent 0123321001233210 etc
	}									// (should be 012321012321 etc)
else
	if(temp->sptr<0)	// if we reach -1, finished the backward run
	{
	temp->sdir=1;           // Start playing forwards again
	temp->sptr=1;		// NEVER have a cyclic loop with 1 frame!
	}
}

/*
 *      anim_random() - random frames
 */

void anim_random(OBJECT *temp)
{
temp->sptr=rndtab[temp->slen][temp->sdir++];
if(temp->sdir == 512)
    temp->sdir=0;
}

/*
 *      Project_Map - Display the map on screen
 */

void Project_Map(int x,int y)
{
int xctr=0;
int yctr=0,ptr,bit,t,ctr,ctr2;

// Find out if we're on schedule to animate everything
Animate = UpdateAnim();

// Tile animation controller

// For each tile...
if(Animate)
    for(ctr=0;ctr<TItot;ctr++)
        {
        // If it animates, update the tiles clock and frame when necessary
        if(TIlist[ctr].slen)
            {
            TIlist[ctr].tick++;
            if(TIlist[ctr].tick>=TIlist[ctr].form->speed)
                {
                tip[ctr]+=TIlist[ctr].sdir;
                if(tip[ctr]>=TIlist[ctr].slen)
                    tip[ctr]=0;
                TIlist[ctr].tick=0;
                }
            }

        // If it scrolls, duplicate the tile in an offscreen buffer and grab
        // it at the appropriate offset to make it appear to scroll
		if(TIlist[ctr].sdx || TIlist[ctr].sdy)
			{
			draw_rle_sprite(tsbscreen,TIlist[ctr].form->seq[tip[ctr]]->image,0,0);
			draw_rle_sprite(tsbscreen,TIlist[ctr].form->seq[tip[ctr]]->image,0,32);
			draw_rle_sprite(tsbscreen,TIlist[ctr].form->seq[tip[ctr]]->image,32,0);
			draw_rle_sprite(tsbscreen,TIlist[ctr].form->seq[tip[ctr]]->image,32,32);
			blit(tsbscreen,tsb[ctr],32-TIlist[ctr].sx,32-TIlist[ctr].sy,0,0,32,32);
			TIlist[ctr].sx = (TIlist[ctr].sx + TIlist[ctr].sdx)&31;
			TIlist[ctr].sy = (TIlist[ctr].sy + TIlist[ctr].sdy)&31;
			}
		}


// Now draw the map

yctr=0;
xctr=0;
ptr=y*curmap->w;
ptr+=x;
bit=curmap->w - VSW;

for(ctr=0;ctr<VSH;ctr++)
	{
	for(ctr2=0;ctr2<VSW;ctr2++)
		{
		t = curmap->physmap[ptr];
		if(t == RANDOM_TILE)
			draw_sprite(gamewin,warning,xctr,yctr);
		else
			{
			if(TIlist[t].sdx || TIlist[t].sdy)
				draw_sprite(gamewin,tsb[t],xctr,yctr);
			else
				draw_rle_sprite(gamewin,TIlist[t].form->seq[tip[t]]->image,xctr,yctr);
			}
		xctr+=32;
		ptr++;
		}
	xctr=0;
	yctr+=32;
	ptr+=bit;
	}
}


/*
 *      Project_sprites- Display the sprites on the map
 */

void Project_Sprites(int x, int y)
{
int cx,cy,vx,vy,yoff;
int startx,starty,ctr;
OBJECT *temp;
SEQ_POOL *project;
OBJECT *postoverlay[MAX_P_OVERLAY];
int post_ovl=0;

roofspr = 0; // No objects on the roof

startx = -VIEWDIST;
starty = -VIEWDIST;

if((x+startx)<0)
	startx=0;
if((y+starty)<0)
	starty=0;

for(vy=starty;vy<VSH;vy++)
	{
	yoff=ytab[y+vy];
	for(vx=startx;vx<VSW;vx++)
		{
		for(temp=curmap->objmap[yoff+x+vx];temp;temp=temp->next)
			{
			if(temp->flags.on)
				{
				if(!temp->flags.invisible)
					{
					cx=vx+x;
					cy=vy+y;

					// Animate the object, continously (or once, if stepped)

					if(temp->form->flags&4)                      // if stepped
						{
						if(temp->flags.stepupdated)              // if updating
							{
							if(temp->form->flags&16)
								anim_random(temp);
							else
								animate[temp->form->flags&3](temp);  // animate once
							temp->flags.stepupdated=0;           // Clear flag
							}
						}
					else
						{
						if(Animate)
							{
							temp->stats->tick++;
							if(temp->stats->tick>=temp->form->speed)
								{
								if(temp->form->flags&16)
									anim_random(temp);
								else
									animate[temp->form->flags&3](temp);  // animate once
								temp->stats->tick=0;
							}
						}
					}

					// Then we plot the object with the CLIP method
					// <<5 multiplies by 32, converting the map coordinates into pixels.

					// Project the sprite
					project = temp->form;
					if(temp->flags.translucent)
						{
//						if((cx-x) >=0 && (cy-y) >=0)
							draw_trans_rle_sprite(gamewin,project->seq[temp->sptr]->image,((cx-x)<<5),((cy-y)<<5));
						}
					else
						if(temp->flags.shadow)
							{
							set_burn_blender(32,32,32,48);
//							if((cx-x) >=0 && (cy-y) >=0)
								draw_trans_rle_sprite(gamewin,project->seq[temp->sptr]->image,((cx-x)<<5),((cy-y)<<5));
							set_trans_blender(0, 0, 0, 128);
							}
						else
							draw_rle_sprite(gamewin,project->seq[temp->sptr]->image,(cx-x)<<5,(cy-y)<<5);

                // If there is an overlay sprite, find out if it is for now
                // or for later.  If it's for now, display it, else queue it

					if(temp->form->overlay /*&& !temp->flags.shadow*/)
						if(temp->flags.overlay)
							{
							if(post_ovl<MAX_P_OVERLAY)
								postoverlay[post_ovl++]=temp;
							}
						else
							draw_rle_sprite(gamewin,project->overlay->image,(cx-x)<<5,(cy-y)<<5);
					}
				else	// Rooftop objects are invisible
					{
					if(temp->flags.overlay)
						if(roofspr<MAX_P_OVERLAY)
							roofsprites[roofspr++]=temp;
					}
				}
			}
		}
	}
// Now, display any post-processed overlays

for(ctr=0;ctr<post_ovl;ctr++)
	{
	temp = postoverlay[ctr];
	draw_rle_sprite(gamewin,temp->form->overlay->image,(temp->x-x)<<5,(temp->y-y)<<5);
	}
}

/*
 *      Project_roof - Display the roof objects on the map
 */

void Project_Roof(int x, int y)
{
int xctr=0,yctr,ptr,bit,ctr,ctr2;
OBJECT *temp;

xctr=0;
yctr=0;

clear(roofwin);	// This must be black for framebuffer overlaying later

ptr=y*curmap->w;
ptr+=x;

bit=curmap->w - VSW;
xctr=0;
for(ctr=0;ctr<VSH;ctr++)
	{
	for(ctr2=0;ctr2<VSW;ctr2++)
		{
		if(curmap->roof[ptr])
			if(curmap->roof[ptr]<RTtot)
				draw_rle_sprite(roofwin,RTlist[curmap->roof[ptr]].image,xctr,yctr);
		xctr+=32;
		ptr++;
		}
	xctr=0;
	yctr+=32;
	ptr+=bit;
	}

// Now, project the rooftop sprites (chimney etc)

for(ctr=0;ctr<roofspr;ctr++)
    {
    temp = roofsprites[ctr];
	draw_rle_sprite(roofwin,temp->form->seq[0]->image,(temp->x-x)<<5,(temp->y-y)<<5);
    }
}

/*
 *      Hide_Unseen - Remove rooms you can't see from the view (like U6)
 */

void Hide_Unseen()
{
int xctr=0,yctr,ctr,ctr2;
int xoff,yoff,xmax,ymax;

xoff = mapx-1;
yoff = mapy-1;
xmax = VSW;
ymax = VSH;

if(player) // If there isn't a player we want to keep the old visibility map
	{
	// First, clear the visibility map
	memset(&vismap[0][0],0,(sizeof(char)*(LMSIZE*LMSIZE)));

	// Now find the tiles that block light

	for(ctr=0;ctr<=ymax;ctr++)
		for(ctr2=0;ctr2<=xmax;ctr2++)
			blmap[ctr2][ctr] = BlocksLight(mapx+ctr2,mapy+ctr);

	vis_floodfill(player->x-mapx,player->y-mapy);
	}

// Block out what we don't want to see

xctr=0;
yctr=0;

xctr=0;
for(ctr=0;ctr<ymax;ctr++)
	{
	for(ctr2=0;ctr2<xmax;ctr2++)
		{
		if(!vismap[ctr2][ctr]&1)
			rectfill(gamewin,xctr,yctr,xctr+32,yctr+32,vis_black);
		xctr+=32;
		}
	xctr=0;
	yctr+=32;
	}

}


void SetDarkness(int d)
{
darklevel=(unsigned char)d;
if(ire_bpp == 8 && darklevel>30)
	darklevel=30;
}



void init_lightmap()
{
int r,x,y,cx,cy;
char filename[1024];
// Load in the light mask

if(!loadfile("sprites/lightmap.cel",filename))
	ithe_panic("Could not find lightmask '%s'",filename);
lightmap=load_lightmap(filename);

// Put the light mask where we can get_sprite it

clear_to_color(darkmap,0);
draw_sprite(darkmap,lightmap,0,0);

// Find size of the light mask in tiles

lmw = lightmap->w/32;
lmh = lightmap->h/32;

// Correct the value (since it rounds down)

if((lmw * 32) < lightmap->w)
    lmw++;
if((lmh * 32) < lightmap->h)
    lmh++;

lmw2=lmw/2;
lmh2=lmh/2;

// Break the lightmap into tiles

for(x=0;x<=lmw;x++)
    for(y=0;y<=lmh;y++)
        {
        lm[x][y]=create_bitmap_ex(8,32,32);
		if(!lm[x][y])
			ithe_panic("Could not create lightmap tile","Out of memory?");
        for(cx=0;cx<32;cx++)
            for(cy=0;cy<32;cy++)
                {
				r = getpixel(darkmap,cx+(x<<5),cy+(y<<5));
				putpixel(lm[x][y],cx,cy,r);
                }
        }

// Now allocate the central hub
// (The inner 3x3 tiles are not subject to solidity checks, so we can just
//  project it always, instead of using tiles and checking.)

lmcore=create_bitmap_ex(8,96,96);
if(!lmcore)
	ithe_panic("Could not create lightmap core","Out of memory?");
for(cx=0;cx<96;cx++)
    for(cy=0;cy<96;cy++)
        {
		r = getpixel(darkmap,cx+32,cy+32);
		putpixel(lmcore,cx,cy,r);
        }
}

/*
 *      render the lighting
 */

void render_lighting(int x, int y)
{
int cx,cy,ex,ey,ctr,nx,ny;
int startx,starty;
OBJECT *temp;
LTAB light[256];
int lights;
unsigned char *lptr;
int offset,id;

clear_to_color(darkmap,darklevel);
memset(&blmap[0][0],0,(sizeof(char)*(LMSIZE*LMSIZE)));

// Find all lightsources

lights=0;

basex=x;
basey=y;

ey = y+VSH+3;
ex = x+VSW+3;
startx = x-3;
starty = y-3;

if(startx<0) startx=0;
if(starty<0) starty=0;
if(ex>=curmap->w)
    ex=curmap->w;
if(ey>=curmap->h)
    ey=curmap->h;

for(cy=starty;cy<ey;cy++)
for(cx=startx;cx<ex;cx++)
    {
    temp = GetRawObjectBase(cx,cy);
    for(;temp;temp=temp->next)              // Traverse the list
        if(temp->flags.on && lights<255)
            if(temp->light)
                {
                // if the object is in the screen region
                if((cx+lmw >= x) && (cx-lmw < ex))
                    if((cy+lmh >= y) && (cy <= ey))
                        {
                        nx = cx-x;
                        ny = cy-y;
                        if(nx+lmw>0 && ny+lmh>0)
                            if(nx-lmw2<VSW && ny-lmh2<VSH)
                                {
                                light[lights].light=temp->light;
                                light[lights].x=nx;
                                light[lights].y=ny;
                                lights++;
                                }
                        }
                }
    }

// Ok, we're ready.  Preserve the default colourmap

oldcmap = color_map;
color_map = &light_cmap;	// Light

// First project any static lighting

offset = curmap->w-VSW;
lptr = &curmap->lightst[(y*curmap->w)+x];

nx=0;ny=0;
for(cy=0;cy<VSH;cy++)
	{
	for(cx=0;cx<VSW;cx++)
		{
		if(*lptr++)
			{
			id=curmap->light[((cy+y)*curmap->w)+cx+x];
			if(id)
				draw_trans_sprite(darkmap,LTlist[id].image,nx,ny);
			}
		nx+=32;
		}
	lptr+=offset;
	nx=0;
	ny+=32;
	}

// Project darksources first

color_map = &dark_cmap; // Anti-light

for(ctr=0;ctr<lights;ctr++)
    if(light[ctr].light<0)
        proj_light(light[ctr].x,light[ctr].y);

// Then project lightsources

color_map = &light_cmap;	// Light

for(ctr=0;ctr<lights;ctr++)
    if(light[ctr].light>0)
        proj_light(light[ctr].x,light[ctr].y);

color_map = oldcmap;

// Now use the darkness map to shade the game window
darken(gamewin,darkmap,VSW32,VSH32);
// And the roof (which is rendered separately to avoid contamination)
darken_s(roofwin,darklevel,VSW32,VSH32);
}

/*
 *      Project a single light source
 */

void proj_light(int x, int y)
{
int xx,yy,cx,cy,ctr;
char dlm[8][8];
int xoffset,yoffset;

#define STARTX 0

xoffset = STARTX;//+(1-rnd(2));
yoffset = STARTX;//+(1-rnd(2));

// Light central hub square

draw_trans_sprite(darkmap,lmcore,((x-1)<<5)+xoffset,((y-1)<<5)+yoffset);

// Clear the already-lit array
memset(dlm,0,64);

cx = x+basex;
cy = y+basey;

for(ctr=0;ctr<8;ctr++)
    {
    // Store for faster access
    xx = (node[ctr][0][0]);
    yy = (node[ctr][0][1]);

    if(!BlocksLight(cx+xx,cy+yy))
        {
        xx = (node[ctr][1][0])+2;
        yy = (node[ctr][1][1])+2;
		if(xx<0 || xx>7)
			ilog_quiet("(1) xx=%d, ctr=%d\n",xx,ctr);
		if(yy<0 || yy>7)
			ilog_quiet("(1) yy=%d, ctr=%d\n",yy,ctr);
        if(!dlm[xx][yy])
            {
			draw_trans_sprite(darkmap,lm[xx][yy],((x+xx-2)<<5)+xoffset,((y+yy-2)<<5)+yoffset);
            dlm[xx][yy]=1;
            }


        xx = (node[ctr][2][0])+2;
        yy = (node[ctr][2][1])+2;
		if(xx<0 || xx>7)
			ilog_quiet("(2) xx=%d, ctr=%d\n",xx,ctr);
		if(yy<0 || yy>7)
			ilog_quiet("(2) yy=%d, ctr=%d\n",yy,ctr);
        if(!dlm[xx][yy])
            {
			draw_trans_sprite(darkmap,lm[xx][yy],((x+xx-2)<<5)+xoffset,((y+yy-2)<<5)+yoffset);
            dlm[xx][yy]=1;
            }

        xx = (node[ctr][3][0])+2;
        yy = (node[ctr][3][1])+2;
		if(xx<0 || xx>7)
			ilog_quiet("(3) xx=%d, ctr=%d\n",xx,ctr);
		if(yy<0 || yy>7)
			ilog_quiet("(3) yy=%d, ctr=%d\n",yy,ctr);

        if(!dlm[xx][yy])
            {
			draw_trans_sprite(darkmap,lm[xx][yy],((x+xx-2)<<5)+xoffset,((y+yy-2)<<5)+yoffset);
            dlm[xx][yy]=1;
            }
        }
    }
}


void vis_floodfill(int x, int y)
{
// Check bounds

if(x<0 || x>VSW)
	return;
if(y<0 || y>VSH)
	return;
/*
if(x<0 || x>(VSW-1))
	return;
if(y<0 || y>(VSH-1))
	return;
*/

// Already done?
if(vismap[x][y])
	return;

// If it blocks light, stop right there
if(blmap[x][y])
	{
	vismap[x][y]=2;
	return;
	}
else
	vismap[x][y]=1;

vis_floodfill(x-1,y-1);
vis_floodfill(x,y-1);
vis_floodfill(x+1,y-1);

vis_floodfill(x-1,y);
vis_floodfill(x+1,y);

vis_floodfill(x-1,y+1);
vis_floodfill(x,y+1);
vis_floodfill(x+1,y+1);
}

/*
void vis_floodfill(int x, int y)
{
// If it blocks light, stop right there
if(blmap[x][y])
	{
	vismap[x][y]=2;
	return;
	}
else
	vismap[x][y]=1;

if(x>0)
	if(!vismap[x-1][y])
		vis_floodfill(x-1,y);
if(x<VSW)
	if(!vismap[x+1][y])
		vis_floodfill(x+1,y);
if(y>0)
	if(!vismap[x][y-1])
		vis_floodfill(x,y-1);
if(y<VSH)
	if(!vismap[x][y+1])
		vis_floodfill(x,y+1);
}
*/
