/*
 *      IRE Game resource loader
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ithelib.h"
#include "oscli.h"
#include "console.h"
#include "media.h"
#include "core.hpp"
#include "textfile.h"
#include "gamedata.h"
#include "sound.h"
#include "init.h"
#include "object.hpp"

#include "pe/pe_api.hpp" // Compiler API

// defines

#define NO_OF_SECTIONS 64       // Up to 64 separate blocks.  Arbitrary.

// Variables

extern "C" int Songs;
extern "C" int Waves;

static struct SectionMap_S              // Map of the blocks in the script
{
    int start;
    int end;
    char name[128];
} sect_map[NO_OF_SECTIONS];

static int no_of_sections;              // Number of blocks in the file
static char Rbuffer[MAX_LINE_LEN];      // Text buffer for LiSP-alikes

static struct TF_S script;              // TextFile loader
static char *compilename;               // Name of file being compiled

// Functions

void LoadResources(char *fn);           // High-level function, compile a file

static void Purify();                   // Clean the text for compiler
static void SeekSect();                 // Map out the blocks

// LiSP and support functions

static void Dump(long lineno,char *error,char *help);  // Report script error

// Process the sections

static void Section(char *name);                   // Section dispatcher
static void PreScan(char *name);                   // pre-parse a section

static void sprites(long start,long finish);       // Compile sprites
static void sequences(long start,long finish);     // Compile animation sequences
static void characters(long start,long finish);    // Compile characters
static void tiles(long start,long finish);         // Compile map tile sequences
static void code(long start,long finish);          // Compile VRM code
static void code2(long start,long finish);         // Compile PE code
static void sounds(long start,long finish);        // Compile sounds
static void music(long start,long finish);         // Compile music
static void rooftiles(long start,long finish);         // Compile music
static void lightmaps(long start,long finish);         // Compile music

static void pre_sprites(long start,long finish);       // Compile sprites
static void pre_sequences(long start,long finish);     // Compile animation sequences
static void pre_characters(long start,long finish);    // Compile characters
static void pre_code(long start,long finish);     		// Compile system scripts

/*
 *      Compile - Compile the script
 */

void LoadResources(char *fn)
{
ilog_printf("Reading game resource file..\n");

TF_init(&script);

ilog_printf("  Loading.. ");
TF_load(&script,fn);
ilog_printf("done.  %ld lines\n",script.lines);

compilename = fn;

ilog_printf("  Purifying.. ");
Purify();
ilog_printf("done.\n");

ilog_printf("  Searching sections.. ");
SeekSect();
ilog_printf("done.  %d sections\n",no_of_sections-1);

ilog_printf("  Parsing.. \n");

// Keep searching for the right sections until we have compiled them all.

PreScan("sprites");
PreScan("sequences");
PreScan("characters");
PreScan("code");

// In order of compilation
Section("sounds");
Section("music");
Section("rooftiles");
Section("lightmaps");

Section("scripts");
// Now we have enough to build the fast lookup tables
Init_Lookups();

Section("code");
Section("sprites");
Section("sequences");
Section("tiles");
Section("characters");

ilog_printf("  Compilation finished.\n");

//script.term();

return;
}

/*
 *      Section - Compile appropriate section, return the number done so far
 */

void Section(char *name)
{
char tbuf[64];
int firstline=0,lastline=0;
char found=0;
int ctr;

for(ctr=0;ctr<no_of_sections;ctr++)
    if(!istricmp(name,sect_map[ctr].name))
        {
        firstline = sect_map[ctr].start;
        lastline = sect_map[ctr].end;
        found=1;
        }

if(!found)
    return;

// Get each end of the section

// Get the section name by stripping off the first element with strrest()

strcpy(tbuf,strrest(script.line[firstline++]));

// Kill off trailing spaces

strstrip(tbuf);

if(!istricmp(name,"code"))
        {
	code(firstline,lastline);
        return;
        }

if(!istricmp(name,"sprites"))
        {
	sprites(firstline,lastline);
        return;
        }

if(!istricmp(name,"sequences"))
        {
	sequences(firstline,lastline);
        return;
        }

if(!istricmp(name,"characters"))
        {
	characters(firstline,lastline);
        return;
        }

if(!istricmp(name,"tiles"))
        {
	tiles(firstline,lastline);
        return;
        }

if(!istricmp(name,"sounds"))
        {
	sounds(firstline,lastline);
        return;
        }

if(!istricmp(name,"music"))
        {
	music(firstline,lastline);
        return;
        }

if(!istricmp(name,"rooftiles"))
	{
	rooftiles(firstline,lastline);
	return;
	}

if(!istricmp(name,"lightmaps"))
	{
	lightmaps(firstline,lastline);
	return;
	}

if(!istricmp(name,"scripts"))
        {
	code2(firstline,lastline);
        return;
        }

ilog_quiet("SC_name='%s'\n",name);
Dump(firstline,"Unknown SECTION type.","Valid types are:\r\nSPRITES  SEQUENCES  TILES  CHARACTERS  CODE  TILES\r\nSOUNDS  MUSIC  ROOFTILES  LIGHTMAPS  SCRIPT");
return;
}

/*
 *      PreScan - gather names for all sections
 */

void PreScan(char *name)
{
char tbuf[64];
int firstline=0,lastline=0;
char found=0;
int ctr;

for(ctr=0;ctr<no_of_sections;ctr++)
    if(!istricmp(name,sect_map[ctr].name))
        {
        firstline = sect_map[ctr].start;
        lastline = sect_map[ctr].end;
        found=1;
        }

if(!found)
    return;

// Get each end of the section

// Get the section name by stripping off the first element with strrest()

strcpy(tbuf,strrest(script.line[firstline++]));

// Kill off trailing spaces

strstrip(tbuf);

if(!istricmp(name,"code"))
        {
		pre_code(firstline,lastline);
        return;
        }

if(!istricmp(name,"sprites"))
        {
	pre_sprites(firstline,lastline);
        return;
        }

if(!istricmp(name,"sequences"))
        {
	pre_sequences(firstline,lastline);
        return;
        }

if(!istricmp(name,"characters"))
        {
	pre_characters(firstline,lastline);
        return;
        }

if(!istricmp(name,"tiles"))
        {
//	pre_tiles(firstline,lastline);
        return;
        }

if(!istricmp(name,"sounds"))
        {
//	pre_sounds(firstline,lastline);
        return;
        }

if(!istricmp(name,"music"))
        {
        return;
        }

if(!istricmp(name,"rooftiles"))
        {
//	pre_rooftiles(firstline,lastline);
        return;
        }

if(!istricmp(name,"lightmaps"))
        {
//	pre_lightmaps(firstline,lastline);
        return;
        }

if(!istricmp(name,"scripts"))
        {
//	pre_code2(firstline,lastline);
        return;
        }

ilog_quiet("PR_name='%s'\n",name);
Dump(firstline,"Unknown SECTION type.","Valid types are:\r\nSPRITES  SEQUENCES  TILES  CHARACTERS  CODE  TILES\r\nSOUNDS  MUSIC  ROOFTILES  LIGHTMAPS  SCRIPT");
return;
}

/* ======================== Section compilers =========================== */

// The sprite list is just a simple list, of format:
//
// IDENTIFIER, FILENAME
// IDENTIFIER, FILENAME
//      :    ,    :
//      :    ,    :
//
// Therefore it is sufficient to just get the first two terms and store them.
//

/*
 *      Pre_Sprites - Get the names for the sprites and allocate space
 */

void pre_sprites(long start,long finish)
{
long ctr,pos,imagedata;
char *Rptr;
char *line;

SPtot = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];        // Get the line
        Rptr = strgetword(line,2);      // If the second word exists,
                                        // then the first word has to as well
        if(Rptr != NOTHING)             // and we assume we have the data
            {
            SPtot++;
            }
        }

// Allocate room on this basis

if(in_editor != 2)
    spr_alloc = SPtot;

ilog_printf("    Creating %d sprite entries ", spr_alloc);

SPlist = (S_POOL *)M_get(sizeof(S_POOL),spr_alloc+1);
ilog_printf("using %d bytes\n", sizeof(S_POOL)*(spr_alloc+1));

// Gather the names and store them

pos = 0;
imagedata=0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];
	Rptr=strfirst(line);                    // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		Rptr=strgetword(line,2);        // Get second term
		if(Rptr!=NOTHING)               // If it exists
			{
                        SPlist[pos].fname = strrest(line);   // Get filename
                        strstrip(SPlist[pos].fname);         // Kill whitespc

                        // Now get the name as a pointer in the text Block

                        SPlist[pos].name=hardfirst(line);
			pos++;  // Go onto the next entry in the sprite list
			}
		}
	}
}

/*
 *      Sprites - Compile the sprites
 */

void sprites(long start,long finish)
{
long ctr,imagedata;

//
//      Most of the grunt-work has already been done, we just need to
//      load in the sprite data for each entry.
//

ilog_printf("    Load sprites..");

Plot(SPtot);
imagedata=0;

for(ctr=0;ctr<SPtot;ctr++)
    imagedata+=Init_Sprite(ctr);

ilog_printf("\n");
ilog_printf("    Done.  %d bytes of image data allocated\n",imagedata);
}

//      Sequences are a more complex format, but we can count the number
//      of sequences in the script by looking for the word 'name:'

/*
 *      Get the names for all the animation sequences
 */

void pre_sequences(long start,long finish)
{
long ctr,pos,spos,tmp;
char *line,*Rptr;
long liststart=0,listend=0;

SQtot = 0;
for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];        // Get the line
	Rptr = strfirst(line);          // First see if it's blank
        if(!istricmp(Rptr,"name"))       // then look for the keyword 'Name'
            SQtot++;                    // Got one
        if(!istricmp(Rptr,"quickname"))  // then look for keyword 'QuickName'
            SQtot++;                    // Got one
        }

// Allocate room on this basis

if(in_editor != 2)
    seq_alloc = SQtot;  // If not in scripter, use original value of seq_alloc

ilog_printf("    Creating %d animation entries ", seq_alloc);

SQlist = (SEQ_POOL *)M_get(sizeof(SEQ_POOL),seq_alloc+1);

ilog_printf("using %d bytes\n", sizeof(SEQ_POOL)*(seq_alloc+1));

// Now, get the name for each entry

pos = 0;
for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];
	Rptr=strfirst(line);                    // Get first term
	if(Rptr!=NOTHING)                       // Main parsing is here
		{
		if(!istricmp(Rptr,"name"))       // Got the first one
			{
			Rptr=strgetword(line,2);        // Get second term
			if(Rptr!=NOTHING)           // Make sure its there
				{
				// Find the offset in the block

				SQlist[pos].name = strrest(line);   // Get name label
				strstrip(SQlist[pos].name);         // Kill whitespace
				pos++;
				}
			}
		if(!istricmp(Rptr,"quickname"))       // Got the first one
			{
			Rptr=strgetword(line,2);        // Get second term
			if(Rptr!=NOTHING)           // Make sure its there
				{
				// Grab the name
				SQlist[pos].name = (char *)M_get(1,strlen(Rptr)+1);
				strcpy(SQlist[pos].name,Rptr);         // Kill whitespace
				pos++;
				}
			}
		}
	}
}

/*
 *      Sequences - Parse the animation sequences
 */

void sequences(long start,long finish)
{
long ctr,pos,spos,tmp,list;
char *line,*Rptr;
long liststart=0,listend=0;
ilog_printf("    Animation sequences");

Plot(SQtot);
pos = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];
	Rptr=strfirst(line);                    // Get first term
	if(Rptr!=NOTHING)                       // Main parsing is here
		{

        // The 'NAME' clause.  Big.

                if(!istricmp(Rptr,"name"))       // Got the first one
                    {
		    Rptr=strgetword(line,2);        // Get second term
	            if(Rptr!=NOTHING)           // Make sure its there
                        {
                        // Find the offset in the block

//                        SQlist[pos].name = strrest(line);   // Get name label
//                        strstrip(SQlist[pos].name);         // Kill whitespace
                        }
                    else
                        Dump(ctr,"NA: This sequence has no name",NULL);

                    SQlist[pos].frames=0;
		    SQlist[pos].speed=0;
		    SQlist[pos].flags=0;
		    SQlist[pos].x=0;
		    SQlist[pos].y=0;
                    SQlist[pos].overlay=NULL;

                    // First, find the boundary of the framelist
                    // so we can calculate the right number of frames

                    liststart=0;
                    listend=0;

                    list = ctr+1;  // Skip the NAME: line
                    do
                        {
                        Rptr = strfirst(script.line[list]);
                        if(!liststart)                  // Don't have start
                            if(!istricmp(Rptr,"framelist:"))
                                liststart = list+1;

                        if(!listend)                    // Don't have end
                            if(!istricmp(Rptr,"end"))
                                listend = list-1;

                        if(!istricmp(Rptr,"name") || list == finish)
                            list = -1;                  // Reached the end

                        list++;
                        } while(list>0);                // list = 0 on exit

                    // Now we should have the boundaries.

                    if(!liststart)
                        Dump(ctr,"NA: This sequence has no FRAMELIST: entry!",SQlist[pos].name);

                    if(!listend)
                        Dump(ctr,"NA: This sequence has no END entry!",SQlist[pos].name);

                    // Good.  Now, we'll count the frames

                    for(list = liststart;list<=listend;list++)
                        {
                        Rptr = strfirst(script.line[list]);
                        if(Rptr != NOTHING)
                            SQlist[pos].frames++;
                        }

                    if(!SQlist[pos].frames)
                        Dump(ctr,"NA: This sequence has no frames at all!",SQlist[pos].name);

                    // Now we have the number of frames in the list

                    SQlist[pos].seq = (S_POOL **)M_get(sizeof(S_POOL *),SQlist[pos].frames+1);

                    // Now we've allocated the space.  We're done I think
                    }

        // The 'FRAMELIST' clause.

                if(!istricmp(Rptr,"framelist:"))       // Got the first one
                    {
                    if(!SQlist[pos].name)
                        Dump(pos,"FL: This sequence has no name",NULL);

                    if(!liststart)
                        Dump(ctr,"FL: This sequence has no FRAMELIST: entry!",SQlist[pos].name);

                    if(!listend)
                        Dump(ctr,"FL: This sequence has no END entry!",SQlist[pos].name);

                    spos=0;

                    for(long list = liststart;list<=listend;list++)
                        {
                        Rptr = strfirst(script.line[list]);
                        if(Rptr != NOTHING)
                            {
                            SQlist[pos].seq[spos] = find_spr(Rptr);
                            if(!SQlist[pos].seq[spos++])
                                Dump(ctr,"FL: Could not find this sprite",Rptr);
                            }
                        }
                    }

        // The 'END' clause.

                if(!istricmp(Rptr,"END"))       // End of the sequence
                    {
                    pos++;
                    Plot(0);
                    }

        // The 'QUICKNAME' clause.

        // QN is a fast-track sequence definition for just one frame.
        // Quicknames are of the form:
        //
        //      quickname  sequence_name sprite_name
        //
        // ..and are therefore much easier for things like map tiles
        // which do not usually have a second frame.

                if(!istricmp(Rptr,"quickname")) // Got the first one
                    {
	            Rptr=strfirst(strrest(strrest(line)));     // Get Third term
	            if(Rptr!=NOTHING)                 // Make sure its there
                        {
                        // For a Quick sequence there is just one frame.
                        SQlist[pos].frames=1;
                        SQlist[pos].seq = (S_POOL **)M_get(sizeof(S_POOL *),SQlist[pos].frames+1);

                        // Now we've allocated the frame, find the sprite
                        SQlist[pos].seq[0] = find_spr(Rptr);

                        // Did it work?
                        if(!SQlist[pos].seq[0])
                            Dump(ctr,"QN: Could not find this sprite",Rptr);

//                        SQlist[pos].name = hardfirst(strrest(line));  // Get name
//                        strstrip(SQlist[pos].name);        // Kill whitespace

//                        boot2("QN: n:%s s:%d\n",SQlist[pos].name,SQlist[pos].seq[0]);

                        }
                    else
                        Dump(ctr,"QN: QuickName sequences must be defined as:\r\n QUICKNAME sequence_name sprite_name",NULL);

		    SQlist[pos].speed=0;
		    SQlist[pos].flags=0;
		    SQlist[pos].x=0;
		    SQlist[pos].y=0;
                    pos++;
                    Plot(0);
                    Rptr=NOTHING;
                    }

        // Flags

                if(!istricmp(Rptr,"pingpong"))       // Animation will rewind
                    {
                    SQlist[pos].flags|=1;
		    if(SQlist[pos].frames<2)
			Dump(ctr,"You must not have a pingpong sequence with only one frame!",NULL);
                    }

                if(!istricmp(Rptr,"loop"))             // Animation repeats
                    SQlist[pos].flags|=2;
                if(!istricmp(Rptr,"looped"))
                    SQlist[pos].flags|=2;
                if(!istricmp(Rptr,"loops"))
                    SQlist[pos].flags|=2;

                if(!istricmp(Rptr,"stepped"))          // Animation is stepped
                    SQlist[pos].flags|=4;

                if(!istricmp(Rptr,"random"))           // Animation is random
                    SQlist[pos].flags|=16;


        // The 'X' clause.

                if(!istricmp(Rptr,"x"))       // X offset
                    {
                    strcpy(Rbuffer,strfirst(strrest(line)));
                    SQlist[pos].x=strgetnumber(Rbuffer);
                    if(!(strisnumber(Rbuffer)))
			Dump(ctr,"The X offset should be a number.",NULL);
                    }


        // The 'Y' clause.

                if(!istricmp(Rptr,"y"))       // Y offset
                    {
                    strcpy(Rbuffer,strfirst(strrest(line)));
                    SQlist[pos].y=strgetnumber(Rbuffer);
                    if(!(strisnumber(Rbuffer)))
			Dump(ctr,"The Y offset should be a number.",NULL);
                    }

        // The 'SPEED' clause.

                if(!istricmp(Rptr,"speed"))       // Animation rate
                    {
                    strcpy(Rbuffer,strfirst(strrest(line)));
                    SQlist[pos].speed=strgetnumber(Rbuffer);
                    if(!(strisnumber(Rbuffer)))
			Dump(ctr,"The speed should be a number.",NULL);
                    }

        // The 'OVERLAY' clause.

                if(!istricmp(Rptr,"overlay"))       // Overlaid frame
                    {
                    strcpy(Rbuffer,strfirst(strrest(line)));
                    SQlist[pos].overlay = find_spr(Rbuffer);
                    // Did it work?
                    if(!SQlist[pos].overlay)
                            Dump(ctr,"OV: Could not find this sprite",Rbuffer);
                    }

        // The 'JUMP_TO' clause, a sequence to switch to when this one ends

                if(!istricmp(Rptr,"jump_to"))       // Sequence to switch to
                    {
                    strcpy(Rbuffer,strfirst(strrest(line)));
                    SQlist[pos].flags|=8;          // Mark that we have a jump

                    // Store the line no we're going to use for post-process
                    // after all sequence definitions have been read in
                    SQlist[pos].jumpto = (SEQ_POOL *)ctr;
                    }

                }
        }

ilog_printf("\n");
ilog_printf("      Aftertouches...\n");

for(ctr=0;ctr<pos;ctr++)
    if(SQlist[ctr].flags&8)
        {
        strcpy(Rbuffer,strfirst(strrest(script.line[(int)SQlist[ctr].jumpto])));
        tmp=getnum4sequence(Rbuffer);
            if(tmp==-1)
                Dump((long)SQlist[ctr].jumpto,"SQJT: Could not find sequence:",Rbuffer);
        SQlist[ctr].jumpto = &SQlist[tmp];
        }
}

/*
 *      Pre-Characters - Get the character names
 */

void pre_characters(long start,long finish)
{
int ctr,pos;
char *line,*Rptr;

// Values for SETSOLID


CHtot = 0;

//      Count the instances of 'Name' to get the quantity of characters.
for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];        // Get the line
	Rptr = strfirst(line);             // First see if it's blank
	if(!istricmp(Rptr,"name"))       // then look for the word 'Name'
		CHtot++;                    // Got one
	}

ilog_printf("    Creating %d characters ", CHtot);

if(in_editor != 2)
    chr_alloc = CHtot;  // If not in scripter, use original value

CHlist = (OBJECT *)M_get(sizeof(OBJECT),chr_alloc+1);
ilog_printf("using %d bytes\n", sizeof(OBJECT)*(chr_alloc+1));

// Extract the names

pos = -1;
for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // Main parsing is here
		{
		if(!istricmp(Rptr,"name"))       // Got the first one
			{
			Rptr=strfirst(strrest(line));     // Get second term
			if(Rptr!=NOTHING)           // Make sure its there
				{
				pos ++;                             // Pre-increment
				CHlist[pos].name = strrest(line);      // Get name label
				strstrip(CHlist[pos].name);         // Kill whitespace
				if(strlen(CHlist[pos].name) > 31)   // Check name length
					Dump(ctr,"This name exceeds 31 letters",CHlist[pos].name);
				}
			}
		}
	}
}

/*
 *      Characters - Compile the characters
 */

void characters(long start,long finish)
{
int ctr,tmp,pos;
char *line,*Rptr;
char something;         // Something is used for error-checking.
                        // It starts as zero, set when a valid line is found.
                        // If there is no valid line, signal an error

unsigned char polarity,blockx,blocky,blockw,blockh;
// Values for SETSOLID


ilog_printf("    Setting up characters");

Plot(CHtot);
pos = -1;

for(ctr=start;ctr<=finish;ctr++)
	{
	something=0;                            // Assume nothing is there
	line = script.line[ctr];
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // Main parsing is here
		{
        // Get the name

		if(!istricmp(Rptr,"name"))       // Got the first one
			{
			Rptr=strfirst(strrest(line));     // Get second term
			if(Rptr!=NOTHING)           // Make sure its there
				{
				// Find the offset in the block

				Plot(0);
				pos ++;                             // Pre-increment
				CHlist[pos].personalname= (char *)M_get(32,1);
				CHlist[pos].maxstats = (STATS *)M_get(sizeof(STATS),1);
				CHlist[pos].stats = (STATS *)M_get(sizeof(STATS),1);
				CHlist[pos].funcs = (FUNCS *)M_get(sizeof(FUNCS),1);
				CHlist[pos].funcs->ucache=-1;
				CHlist[pos].funcs->tcache=-1;
				CHlist[pos].funcs->kcache=-1;
				CHlist[pos].funcs->lcache=-1;
				CHlist[pos].funcs->scache=-1;
				CHlist[pos].funcs->hcache=-1;
				CHlist[pos].funcs->icache=-1;
				CHlist[pos].funcs->dirty=0;
				CHlist[pos].funcs->wcache=-1;

				CHlist[pos].vblock[BLK_X]=0;   // Partly-solid object
				CHlist[pos].vblock[BLK_Y]=0;   // Solidity offset
				CHlist[pos].vblock[BLK_W]=0;   // and size
				CHlist[pos].vblock[BLK_H]=0;
				CHlist[pos].hblock[BLK_X]=0;   // Partly-solid object
				CHlist[pos].hblock[BLK_Y]=0;   // Solidity offset
				CHlist[pos].hblock[BLK_W]=0;   // and size
				CHlist[pos].hblock[BLK_H]=0;
//				CHlist[pos].behave = -1;            // No behaviour
//				CHlist[pos].stats->oldbehave = -1;  // No behaviour
				CHlist[pos].stats->tick=0;
				CHlist[pos].stats->owner=NULL;
				CHlist[pos].stats->npctalk=NULL;
				CHlist[pos].target=NULL;
				CHlist[pos].activity=-1;

				strcpy(CHlist[pos].personalname,"-");
				CHlist[pos].desc = "No description";
				CHlist[pos].shortdesc = CHlist[pos].desc;
				// By default the thing is resurrected to be itself
				strcpy(CHlist[pos].funcs->resurrect,CHlist[pos].name);
				something=1;                        // Got something
				}
			else
				Dump(ctr,"NA: This character has no name",NULL);
			}

        // Sanity Check
		if(pos == -1)
			Dump(ctr,"NA: The character's name must come first",NULL);

        // Get the vital statistics

		if(!istricmp(Rptr,"hp"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's health must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->hp=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_hp"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's maximum health must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->hp=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"dexterity"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's dexterity must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->dex=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_dexterity"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's maximum dexterity must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->dex=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"strength"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's strength must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->str=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_strength"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's maximum strength must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->str=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"intelligence"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's intelligence must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->intel=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_intelligence"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's maximum intelligence must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->intel=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"weight"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The character's weight must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->weight=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"damage"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The damage it causes must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->damage=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_damage")) // is this needed?
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The maximum damage it causes must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->damage=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"karma"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The karma it adds/subtracts must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->karma=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"bulk"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The object's bulk must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].stats->bulk=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"max_bulk"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The object's bulk capacity must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
			CHlist[pos].maxstats->bulk=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"left"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			tmp=getnum4sequence(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find sequence:",Rbuffer);
			CHlist[pos].dir[CHAR_L]=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"right"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			tmp=getnum4sequence(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find sequence:",Rbuffer);
			CHlist[pos].dir[CHAR_R]=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"up"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			tmp=getnum4sequence(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find sequence:",Rbuffer);
			CHlist[pos].dir[CHAR_U]=tmp;
			something=1;                        // Got something
			}

		if(!istricmp(Rptr,"down"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			tmp=getnum4sequence(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find sequence:",Rbuffer);
			CHlist[pos].dir[CHAR_D]=tmp;
			something=1;                        // Got something
			}

        // Get a text description

		if(!istricmp(Rptr,"description"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));

			if(Rbuffer[0] == '\"')  // Quoted text is direct
				{
				CHlist[pos].desc = strchr(line,'\"'); // start
				CHlist[pos].desc++;            // skip the quote
				Rptr = strrchr(line,'\"');     // Find last quote
				if(Rptr)
					*Rptr = 0;                 // blow quote away
				}
			else                    // Look up unquoted text
				Dump(ctr,"Invalid string",Rbuffer);
			something=1;                    // Got something
			}

        // Get the short description

		if(!istricmp(Rptr,"short"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));

                        if(Rbuffer[0] == '\"')  // Quoted text is direct
                            {
                            CHlist[pos].shortdesc = strchr(line,'\"'); // start
                            CHlist[pos].shortdesc++;       // skip the quote
                            Rptr = strrchr(line,'\"');     // Find last quote
                            if(Rptr)
                                *Rptr = 0;                 // blow quote away
                            }
                        else                    // Look up unquoted text
                            Dump(ctr,"Invalid string",Rbuffer);
                        something=1;                    // Got something
			}

        // Get the character's USE function

		if(!istricmp(Rptr,"IfUsed"))
			{
			strcpy(CHlist[pos].funcs->use,strfirst(strrest(line)));
			something=1;                    // Got something
			}

        // Get the character's behaviour function

		if(!istricmp(Rptr,"behave") || !istricmp(Rptr,"behaviour")
		   || !istricmp(Rptr,"behavior"))
			{
			if(CHlist[pos].activity != -1)
				Dump(ctr,"Multiple initial behaviours not supported",NULL);
			CHlist[pos].activity = getnum4PE(strfirst(strrest(line)));
			if(CHlist[pos].activity == -1)
				Dump(ctr,"Cannot find script function",strfirst(strrest(line)));
/*                        if(!(strisnumber(strfirst(strrest(line)))))
			    Dump(ctr,"Format is: BEHAVIOUR <priority number> <VRM name>",NULL);
			tmp = strgetnumber(strfirst(strrest(line)));

            if(!AddGoalTemplate(&CHlist[pos],tmp,strrest(strrest(line)),NULL))
                            Dump(ctr,"Can't add goal, don't know why",NULL);
                        CHlist[pos].goals = (GOAL *)M_get(1,sizeof(GOAL));
                        strcpy(CHlist[pos].goals->vrm,first(strrest(strrest(line))));
                        CHlist[pos].goals->priority = tmp;
                        CHlist[pos].goals->next = NULL;
*/
                        something=1;                    // Got something
			}

        // Get the character's STAND function

		if(!istricmp(Rptr,"IfTriggered"))
			{
                        strcpy(CHlist[pos].funcs->stand,strfirst(strrest(line)));
                        CHlist[pos].flags.trigger = 1;

                        if(!in_editor)
							{
							CHlist[pos].activity = getnum4PE("if_triggered");
							if(CHlist[pos].activity == -1)
								Dump(ctr,"Cannot find script function if_triggered",NULL);
							}
//                            AddGoalTemplate(&CHlist[pos],-127,"iftriggered",NULL);

                        something=1;                    // Got something
			}

        // Get the character's KILLED function

		if(!istricmp(Rptr,"IfKilled") || !istricmp(Rptr,"IfDead"))
			{
                        strcpy(CHlist[pos].funcs->kill,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

                // Get the character's LOOK function

		if(!istricmp(Rptr,"IfLooked") || !istricmp(Rptr,"IfLookedAt")
                ||!istricmp(Rptr,"IfLook"))
			{
                        strcpy(CHlist[pos].funcs->look,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

        // Get the character's HURT function

		if(!istricmp(Rptr,"IfHurt") || !istricmp(Rptr,"IfDamaged"))
			{
                        strcpy(CHlist[pos].funcs->hurt,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

        // Get the character's INIT function

		if(!istricmp(Rptr,"Init") || !istricmp(Rptr,"OnInit"))
			{
                        strcpy(CHlist[pos].funcs->init,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

        // Get the character's WIELD function

		if(!istricmp(Rptr,"IfWield") || !istricmp(Rptr,"IfWielded"))
			{
                        strcpy(CHlist[pos].funcs->wield,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

        // Get the character's resurrection object, if special

		if(!istricmp(Rptr,"resurrect") || !istricmp(Rptr,"resurrect_to")
                || !istricmp(Rptr,"resurrect_as"))
			{
                        strcpy(CHlist[pos].funcs->resurrect,strfirst(strrest(line)));
                        something=1;                    // Got something
			}

		if(!istricmp(Rptr,"no_resurrect"))
                        {
                        strcpy(CHlist[pos].funcs->resurrect,"-");
                        something=1;                    // Got something
                        }

        // Get the character's light level
        // Invert it to bring in line with my bizarre light physics

		if(!istricmp(Rptr,"light"))
			{
                        CHlist[pos].light = strgetnumber(strfirst(strrest(line)));
                        something=1;                    // Got something
			}

        // Is the character solid only in places?
        // If so, find out which part is solid and how big it is

		if(!istricmp(Rptr,"setsolid")
                 ||!istricmp(Rptr,"subsolid"))
			{
//                        CHlist[pos].flags.solid = 1;  // Make sure it's solid
                        polarity=0;
                        strcpy(Rbuffer,strgetword(line,2));
			if(!istricmp(Rbuffer,"H"))
                                polarity=1;
			if(!istricmp(Rbuffer,"V"))
                                polarity=2;
                        if(!polarity)
                                Dump(ctr,"\r\nPartly-solid objects must be declared with a polarity and four numbers:\r\nSETSOLID [H or V] <x offset> <y offset> <width> <height>",Rbuffer);

                        strcpy(Rbuffer,strgetword(line,3));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nPartly-solid objects must be declared with a polarity and four numbers:\r\nSETSOLID [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockx= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,4));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nPartly-solid objects must be declared with a polarity and four numbers:\r\nSETSOLID [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blocky= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,5));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nPartly-solid objects must be declared with a polarity and four numbers:\r\nSETSOLID [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockw= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,6));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nPartly-solid objects must be declared with a polarity and four numbers:\r\nSETSOLID [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockh= strgetnumber(Rbuffer);

                        if(polarity == 1)
                            {
                            CHlist[pos].hblock[BLK_X] = blockx;
                            CHlist[pos].hblock[BLK_Y] = blocky;
                            CHlist[pos].hblock[BLK_W] = blockw;
                            CHlist[pos].hblock[BLK_H] = blockh;
                            }
                        else
                            {
                            CHlist[pos].vblock[BLK_X] = blockx;
                            CHlist[pos].vblock[BLK_Y] = blocky;
                            CHlist[pos].vblock[BLK_W] = blockw;
                            CHlist[pos].vblock[BLK_H] = blockh;
                            }
                        something=1;                    // Got something
			}

        // Is the character smaller than it's physical appearance?
        // If so, find out which part is active and how big it is

		if(!istricmp(Rptr,"setarea")
                 ||!istricmp(Rptr,"setactive")
                 ||!istricmp(Rptr,"setactivearea"))
			{
                        polarity=0;
                        strcpy(Rbuffer,strgetword(line,2));
			if(!istricmp(Rbuffer,"H"))
                                polarity=1;
			if(!istricmp(Rbuffer,"V"))
                                polarity=2;
                        if(!polarity)
                                Dump(ctr,"\r\nObjects with Active Areas must be declared with a polarity and four numbers:\r\nSetActiveArea [H or V] <x offset> <y offset> <width> <height>",Rbuffer);

                        strcpy(Rbuffer,strgetword(line,3));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nObjects with Active Areas must be declared with a polarity and four numbers:\r\nSetActiveArea [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockx= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,4));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nObjects with Active Areas must be declared with a polarity and four numbers:\r\nSetActiveArea [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blocky= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,5));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nObjects with Active Areas must be declared with a polarity and four numbers:\r\nSetActiveArea [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockw= strgetnumber(Rbuffer);

                        strcpy(Rbuffer,strgetword(line,6));
			if(!strisnumber(Rbuffer))
                                Dump(ctr,"\r\nObjects with Active Areas must be declared with a polarity and four numbers:\r\nSetActiveArea [H or V] <x offset> <y offset> <width> <height>",Rbuffer);
                        blockh= strgetnumber(Rbuffer);

                        if(polarity == 1)
                            {
                            CHlist[pos].harea[BLK_X] = blockx;
                            CHlist[pos].harea[BLK_Y] = blocky;
                            CHlist[pos].harea[BLK_W] = blockw;
                            CHlist[pos].harea[BLK_H] = blockh;
                            }
                        else
                            {
                            CHlist[pos].varea[BLK_X] = blockx;
                            CHlist[pos].varea[BLK_Y] = blocky;
                            CHlist[pos].varea[BLK_W] = blockw;
                            CHlist[pos].varea[BLK_H] = blockh;
                            }
                        something=1;                    // Got something
			}

        // Get the flags

                if(!istricmp(Rptr,"solid"))             // Is it solid?
                    {
                    CHlist[pos].flags.solid = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"will_open"))         // Isn't always solid
                    {
                    CHlist[pos].flags.willopen= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"ranged_weapon") ||
                   !istricmp(Rptr,"range_weapon"))      // Is the weapon used
                    {                                  // at a distance or
                    CHlist[pos].flags.rangeweapon = 1; // close-quarters?
                    something=1;
                    }

                if(!istricmp(Rptr,"blocklight") ||
                   !istricmp(Rptr,"blockslight"))        //Is it solid?
                    {
                    CHlist[pos].flags.blocklight= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"invisible"))             // Is it invisible?
                    {
                    CHlist[pos].flags.invisible = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"translucent"))           // Is it translucent?
                    {
                    CHlist[pos].flags.translucent= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"post_overlay"))     // Are overlays postprojected?
                    {
                    CHlist[pos].flags.overlay= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"fixed"))             // Can't move/get it?
                    {
                    CHlist[pos].flags.fixed = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"container"))         // Is a bag or chest?
                    {
                    CHlist[pos].flags.container= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"wielded"))           // Can it be wielded?
                    {
                    CHlist[pos].flags.wield= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"tabletop"))           // Can you drop things
                    {                                   // On this solid object
                    CHlist[pos].flags.tabletop= 1;
                    CHlist[pos].flags.solid = 1;        // Make sure it's solid
                    something=1;
                    }

                if(!istricmp(Rptr,"spikeproof"))         // unaffected by triggers
                    {
                    CHlist[pos].flags.spikeproof= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"person"))          // can't be stuffed in sacks etc
                    {
                    CHlist[pos].flags.person= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"quantity"))        // Can it be a pile?
                    {
                    CHlist[pos].flags.quantity= 1;

                    strcpy(Rbuffer,strfirst(strrest(line)));
                    tmp=strgetnumber(Rbuffer);
                    if(strisnumber(Rbuffer))
                        CHlist[pos].stats->quantity=tmp;
                    else
                        CHlist[pos].stats->quantity=1;
                    something=1;
                    }

                if(!istricmp(Rptr,"boat"))            // Is it a boat?
                    {
                    CHlist[pos].flags.watery= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"shadow"))           // Is it like a shadow?
                    {
                    CHlist[pos].flags.shadow = 1;
                    something=1;
                    }

        // Get the character's speech file

		if(!istricmp(Rptr,"Conversation"))
			{
			strcpy(CHlist[pos].funcs->talk,strfirst(strrest(line)));
			CHlist[pos].funcs->tcache=1;
			something=1;                    // Got something
			}

        // Get the character's default contents

		if(!istricmp(Rptr,"contains"))
			{
                        if(CHlist[pos].funcs->contents<8)
                            strcpy(CHlist[pos].funcs->contains[CHlist[pos].funcs->contents++],strfirst(strrest(line)));
                        else
                            Dump(ctr,"This character has more than 8 initial items",Rbuffer);
                        something=1;                    // Got something
			}

                if(!istricmp(Rptr,"fragile"))             // Is it fragile?
                    {
                    CHlist[pos].flags.fragile = 1;
                    something=1;
                    }

				if(!istricmp(Rptr,"onroof")||!istricmp(Rptr,"on_roof")||
					!istricmp(Rptr,"chimney"))             // On the roof?
                    {
                    CHlist[pos].flags.invisible= 1;
                    CHlist[pos].flags.overlay = 1;
                    something=1;
                    }

// NPC flags

                if(!istricmp(Rptr,"male"))           // Is it male? (default)
                    {
                    CHlist[pos].stats->npcflags.female= 0;
                    something=1;
                    }

                if(!istricmp(Rptr,"female"))           // Is it female?
                    {
                    CHlist[pos].stats->npcflags.female = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"know_name"))      // Player know their name
                    {
                    CHlist[pos].stats->npcflags.know_name = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"is_hero"))      // Player know their name
                    {
                    CHlist[pos].stats->npcflags.is_hero = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"cant_eat"))      // Can't eat or drink
                    {
                    CHlist[pos].stats->npcflags.cant_eat = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"cant_drink"))    // Can't eat or drink
                    {
                    CHlist[pos].stats->npcflags.cant_eat = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"wont_shut_doors")) // Can't shut doors
                    {
                    CHlist[pos].stats->npcflags.no_shutdoor = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"link_to_npc")
				|| !istricmp(Rptr,"link_to_other_npc")
				|| !istricmp(Rptr,"link_to_another_npc")
				|| !istricmp(Rptr,"symlink"))
                    {
                    CHlist[pos].stats->npcflags.symlink= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"biological")) // Flesh and blood
                    {
                    CHlist[pos].stats->npcflags.biological= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"guard")) // It's the cops!
                    {
                    CHlist[pos].stats->npcflags.guard= 1;
                    something=1;
                    }


// Finish

	        if(!something && !skip_unk)
			Dump(ctr,"Unknown keyword:",Rptr);

                }
        }

/*
		if(!istricmp(Rbuffer,"overlay"))
			{
			strcpy(Rbuffer,first(rest(Sbuffer[ctr])));
			tmp=getnum4char(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find character called:",Rbuffer);
			for(ctr2=0;ctr2<6;ctr2++)
				if(CHlist[CHtot].overlay[ctr2]==0 && tmp)
					{
					CHlist[CHtot].overlay[ctr2]=tmp;
//					printf("Entered overlay %x to character %s\n",tmp,CHlist[CHtot].name);
					tmp=0;
					}
				if(tmp!=0)
					Dump(ctr,"This character already has 6 overlays.",Rbuffer);
			something=1;
			}
*/

ilog_printf("\n");
}

/*
 *      Tiles - Compile the tiles
 *              This compiler is based on characters()
 */

void tiles(long start,long finish)
{
int ctr,tmp,pos;
char *line,*Rptr;
char something;         // Something is used for error-checking.
                        // It starts as zero, set when a valid line is found.
                        // If there is no valid line, signal an error

TItot = 0;

//      Count the instances of 'Name' to get the quantity of tiles.

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];        // Get the line
	Rptr = strfirst(line);             // First see if it's blank
        if(!istricmp(Rptr,"name"))       // then look for the word 'Name'
            TItot++;                    // Got one
        }

ilog_printf("    Creating %d tiles ", TItot);

// Allocate room on this basis
if(in_editor != 2)
    til_alloc = TItot;  // If not in scripter, use original value

TIlist = (TILE *)M_get(sizeof(TILE),til_alloc+1);

ilog_printf("using %d bytes\n", sizeof(TILE)*(til_alloc+1));

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

ilog_printf("    Loading tiles..");

Plot(TItot);
pos = -1;

for(ctr=start;ctr<=finish;ctr++)
	{
        something=0;                            // Assume nothing is there
        line = script.line[ctr];
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // Main parsing is here
		{

        // Get the name

                if(!istricmp(Rptr,"name"))       // Got the first one
                    {
	            Rptr=strfirst(strrest(line));     // Get second term
	            if(Rptr!=NOTHING)           // Make sure its there
                        {
                        // Find the offset in the block

                        pos ++;                             // Pre-increment
                        Plot(0);
                        TIlist[pos].name = strrest(line);      // Get name label

                        strstrip(TIlist[pos].name);            // Kill whitespace
                        if(strlen(TIlist[pos].name) > 31)   // Check name length
                            Dump(ctr,"This name exceeds 31 letters",TIlist[pos].name);
			something=1;                        // Got something
                        }
                    else
                        Dump(ctr,"NA: This tile has no name",NULL);
                    }

        // Sanity Check

                if(pos == -1)
                        Dump(ctr,"NA: The tile's name must come first",NULL);

        // Get the sequence of the tile

		if(!istricmp(Rptr,"sequence"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			tmp=getnum4sequence(Rbuffer);
			if(tmp==-1)
				Dump(ctr,"Could not find sequence:",Rbuffer);

                        // Get seq. directly
                        TIlist[pos].form=&SQlist[tmp];
                        // Get name of sequence too
                        strcpy(TIlist[pos].seqname,strfirst(strrest(line)));

                        TIlist[pos].slen = SQlist[tmp].frames;
                        TIlist[pos].sdir = 1;
                        TIlist[pos].sptr = 0;

			something=1;                    // Got something
			}

        // Get the description

		if(!istricmp(Rptr,"description"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));

                        if(Rbuffer[0] == '\"')  // Quoted text is direct
                            {
                            TIlist[pos].desc = strchr(line,'\"'); // start
                            TIlist[pos].desc++;            // skip the quote
                            Rptr = strrchr(line,'\"');     // Find last quote
                            if(Rptr)
                                *Rptr = 0;                 // blow quote away
                            }
                        else                    // Look up unquoted text
                            Dump(ctr,"Invalid string",Rbuffer);
                        something=1;                    // Got something
			}

        // Get the flags

                if(!istricmp(Rptr,"solid"))             // Is it solid?
                    {
                    TIlist[pos].flags.solid = 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"blocklight") ||
                   !istricmp(Rptr,"blockslight"))        //Is it solid?
                    {
                    TIlist[pos].flags.blocklight= 1;
                    something=1;
                    }

                if(!istricmp(Rptr,"water"))             // Is it water?
                    {
                    TIlist[pos].flags.watery= 1;
                    something=1;
                    }

		if(!istricmp(Rptr,"scroll"))
			{
			strcpy(Rbuffer,strfirst(strrest(line)));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The scrolling X coordinate must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
                        if(!in_editor)                  // Only for game
                            if(tmp<0)
                                 tmp=32+tmp;
			TIlist[pos].sdx=tmp;

			strcpy(Rbuffer,strfirst(strrest(strrest(line))));
			if(!strisnumber(Rbuffer))
				Dump(ctr,"The scrolling Y coordinate must be a number.",NULL);
			tmp=strgetnumber(Rbuffer);
                        if(!in_editor)                  // Only for game
                            if(tmp<0)
                                 tmp=32+tmp;
			TIlist[pos].sdy=tmp;

			something=1;                        // Got something
			}

// Finish

	        if(!something && !skip_unk)
			Dump(ctr,"Unknown keyword:",Rbuffer);

                }
        }
ilog_printf("\n");
}

/*
 *      Pre-Code - Get the VRM names
 */

void pre_code(long start,long finish)
{
long ctr,pos,imagedata;
char *Rptr;
char *line;

COtot = 0;

for(ctr= start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];			// Get the line
	strupr(line);						// convert to upper-case
	Rptr = strfirst(strrest(line));		// If the second word exists,
										// then the first word has to as well
	if(Rptr != NOTHING)					// and we assume we have the data
		COtot++;
	}

// Allocate room on this basis

ilog_printf("    Creating %d scripts ",COtot);

if(in_editor != 2)
    vrm_alloc = COtot;  // If not in scripter, use original value
COlist = (VRM *)M_get(sizeof(VRM),vrm_alloc+1);
ilog_printf("using %d bytes\n", sizeof(COlist)*(vrm_alloc+1));

// Gather the names and store them

pos = 0;
imagedata=0;

for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];
	Rptr=strfirst(line);                    // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		Rptr=strgetword(line,2);        // Get second term
		if(Rptr!=NOTHING)               // If it exists
			{
			COlist[pos].fname = strrest(line);      // Get filename
			strstrip(COlist[pos].fname);            // Kill whitespc
			// Now get the name as a pointer in the text Block
			COlist[pos].name= hardfirst(line);
			COlist[pos].count=0;
			pos++;  // Go onto the next entry in the sprite list
			}
		}
	}
}


/*
 *      VRM Code - Load in the VRMS
 *      this compiler is based on sprites()
 */

void code(long start,long finish)
{
long ctr,pos;


ilog_printf("    Loading System scripts");
Plot(COtot);
for(ctr=0;ctr<COtot;ctr++)
	Init_VRM(ctr);

ilog_printf("\n");
}

/*
 *      PE Code - Load in the PissEasy script files
 */

void code2(long start,long finish)
{
long ctr,ctr2,pos;
char *Rptr;
char *line;
OBJCODE *pe;

// Temporary structures for the PissEasy compiler
int numfiles,func;
int *filefuncs;


// The PE list is just a straight list of files.
// We just get the first term and store it.
//

// First, see how many PE files we have.  This != functions

ilog_printf("    Game scripts...");

numfiles = 0;

for(ctr= start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];        // Get the line
        strupr(line);                   // convert to upper-case
	if(strfirst(line) != NOTHING)  // Anything there?
            numfiles++;
        }

// Allocate room for the list of files to be compiled

if(in_editor != 2)
    pef_alloc = numfiles;  // If not in scripter, use original value

pe_files = (char **)M_get(sizeof(char *),pef_alloc+1);
filefuncs = (int *)M_get(sizeof(int),numfiles+1);

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

pos = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];
	Rptr=strfirst(line);                    // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
                pe_files[pos] = hardfirst(line);  // Get filename
                strstrip(pe_files[pos]);
                pos++;  // Go onto the next entry in the VRM list
		}
	}

//ilog_quiet("Used %d/%d entries\n",pos,pef_alloc+1);

compiler_init();

// Gather number of functions so we can allocate space

PEtot=0;
for(pos=0;pos<numfiles;pos++)
    {
    compiler_openfile(pe_files[pos]);
    ctr = compiler_parse();
    filefuncs[pos]=ctr;
    PEtot+=ctr;
    compiler_closefile();
    }

PElist = (PEM *)M_get(PEtot+1,sizeof(PEM));

Plot(PEtot);

func=0;

for(pos=0;pos<numfiles;pos++)
    {
    compiler_openfile(pe_files[pos]);

    // Allocate space for compiled code, length of code, function names

    // (Function names are pointers to the keyword structure and are released
    //  from memory when compiler_term is called)

	if(filefuncs[pos])
		{
		pe = (OBJCODE *)M_get(filefuncs[pos],sizeof(OBJCODE));
		// Compile this source file
		compiler_build(pe);

		// Deal with all the functions in each file

		for(ctr=0;ctr<filefuncs[pos];ctr++)
			{
			// Build all the functions, fill out the output data
			Plot(0);

			// Store the compiled stuff for later
			PElist[func].size = pe[ctr].size;
			PElist[func].codelen = pe[ctr].codelen;
			PElist[func].name = (char *)M_get(1,strlen(pe[ctr].name)+1);
			strcpy(PElist[func].name,pe[ctr].name);
			PElist[func].file = pe_files[pos];
			PElist[func].hidden = pe[ctr].hidden; // hide local functions

			PElist[func].code = (char *)M_get(pe[ctr].size,sizeof(unsigned char));
			memcpy(PElist[func].code,pe[ctr].code,pe[ctr].size);
			Init_PE(func); // Register any system functions
			func++;
			M_free(pe[ctr].code);
			}
		M_free(pe);
		}

	// Ok, we're done with that file
	compiler_closefile();
	}

// We're finished

compiler_term();
ilog_printf("\n");
}

/*
 *      Sounds - Load in the Sounds
 *               this compiler is based on sprites()
 */

void sounds(long start,long finish)
{
long ctr,pos;
char *Rptr;
char *line;

// The WAVE list is just a simple list, of format:
//
// NAME, FILENAME
// NAME, FILENAME
//   : ,    :
//   : ,    :
//
// Therefore it is sufficient to just get the first two terms and store them.
//

// First, see how many sounds we have.

Waves = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];        // Get the line
	Rptr = strfirst(strrest(line)); // If the second word exists,
                                    // then the first word has to as well
	if(Rptr != NOTHING)             // and we assume we have the data
		Waves++;
	}

ilog_printf("    Creating %d sound entries", Waves);

// Allocate room on this basis

if(in_editor != 2)
    wav_alloc = Waves;  // If not in scripter, use original value

//wavtab = (SMTab *)M_get(sizeof(SMTab),Waves+1);

wavtab = (SMTab *)M_get(sizeof(SMTab),wav_alloc+1);
ilog_printf(" %d bytes allocated\n", sizeof(SMTab)*(wav_alloc+1));

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

pos = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		Rptr=strfirst(strrest(strrest(line)));   // Get third term
		if(Rptr!=NOTHING)               // If it exists
			if(!istricmp("nodrift",strfirst(Rptr)))
				wavtab[pos].nodrift = 1;

		Rptr=strfirst(strrest(line));         // Get second term
		if(Rptr!=NOTHING)               // If it exists
			{
                        wavtab[pos].fname = hardfirst(strrest(line));      // Get filename
                        strstrip(wavtab[pos].fname);            // Kill whitespc

                        // Now get the name as a pointer in the text Block
                        wavtab[pos].name= hardfirst(line);
			pos++;  // Go onto the next entry in the WAV list
			}
		}
	}
}

/*
 *      Music - Load in the music
 *              this compiler is based on sprites()
 */

void music(long start,long finish)
{
long ctr,pos;
char *Rptr;
char *line;

// The MUSIC list is just a simple list, of format:
//
// NAME, FILENAME
// NAME, FILENAME
//   : ,    :
//   : ,    :
//
// Therefore it is sufficient to just get the first two terms and store them.
//

// First, see how many mods we have.

//if(in_editor) return;

Songs = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];        // Get the line
	Rptr = strfirst(strrest(line));       // If the second word exists,
                                        // then the first word has to as well
        if(Rptr != NOTHING)             // and we assume we have the data
            Songs++;
        }

ilog_printf("    Creating %d music entries", Songs);

// Allocate room on this basis

if(in_editor != 2)
    mus_alloc = Songs;  // If not in scripter, use original value
mustab = (SMTab *)M_get(sizeof(SMTab),mus_alloc+1);

ilog_printf(" using %d bytes\n", sizeof(SMTab)*(mus_alloc+1));

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

pos = 0;

for(ctr=start;ctr<=finish;ctr++)
	{
        line = script.line[ctr];
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		Rptr=strfirst(strrest(strrest(line)));   // Get third term
		if(Rptr!=NOTHING)               // If it exists
			if(!istricmp("loop",strfirst(Rptr)))
				mustab[pos].loopsong = 1;
		Rptr=strfirst(strrest(line));         // Get second term
		if(Rptr!=NOTHING)               // If it exists
			{
                        mustab[pos].fname = hardfirst(strrest(line));      // Get filename
                        strstrip(mustab[pos].fname);            // Kill whitespc
			strlwr(mustab[pos].fname);

                        // Now get the name as a pointer in the text Block

                        mustab[pos].name= hardfirst(line);
			pos++;  // Go onto the next entry in the MOD list
			}
		}
	}
}

/*
 *      rooftiles - Load in sprites to be used as roof tiles
 */

void rooftiles(long start,long finish)
{
long ctr,pos,imagedata;
char *Rptr;
char *line;

// The rooftiles list is just a single list of sprite files:
//
// FILENAME
// FILENAME
//      :
//      :
//
// Therefore it is sufficient to just get the first term and store it.
//

// First, see how many sprites we have.

RTtot = 1; // 0 is special (no roof tile)

for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];        // Get the line
	Rptr = strfirst(line);             // If the second word exists,
                                        // then the first word has to as well
	if(Rptr != NOTHING)             // and we assume we have the data
		RTtot++;
	}

ilog_printf("    Rooftops..", RTtot);

// Allocate room on this basis

if(in_editor != 2)
    rft_alloc = RTtot;
/*
else
    ilog_printf("      Allocating space for %d entries\n", rft_alloc);
*/

RTlist = (S_POOL *)M_get(sizeof(S_POOL),rft_alloc+1);
//ilog_printf("      %d bytes allocated\n", sizeof(S_POOL)*(rft_alloc+1));

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

Plot(RTtot);
pos = 1; // 0 is special (no roof tile)
imagedata=0;

for(ctr=start;ctr<=finish;ctr++)
	{
    // Stuff an integer 1 into a pointer slot to use as a flag later
    RTlist[pos].name=(char *)1L;
    line = script.line[ctr];
    Rptr=strfirst(strrest(line));
    if(!istricmp(Rptr,"stand_under"))
    	RTlist[pos].name=(char *)0L; //..and again..
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		RTlist[pos].fname = hardfirst(line);     // Get filename
		imagedata += Init_RoofTile(pos);
		pos++;  // Go onto the next entry in the sprite list
		}
	}

ilog_quiet("Finished\n");
ilog_printf("\n");
}


/*
 *      lightmaps - Load in sprites to be used as light maps
 */

void lightmaps(long start,long finish)
{
long ctr,pos,imagedata;
char *Rptr;
char *line;

// The lightmaps list is just a single list of sprite files:
//
// FILENAME
// FILENAME
//      :
//      :
//
// Therefore it is sufficient to just get the first term and store it.
//

// First, see how many sprites we have.

LTtot = 1; // 0 is special (no lightmap)

for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];			// Get the line
	Rptr = strfirst(line);
	if(Rptr != NOTHING)
		LTtot++;
	}

ilog_printf("    Lightmaps..", LTtot);

// Allocate room on this basis

if(in_editor != 2)
    lgt_alloc = LTtot;
/*
else
    ilog_printf("      Allocating space for %d entries\n", rft_alloc);
*/

LTlist = (L_POOL *)M_get(sizeof(L_POOL),lgt_alloc+1);
//ilog_printf("      %d bytes allocated\n", sizeof(S_POOL)*(rft_alloc+1));

// Now, process the data.  We will never parse this again, so it can be done
// destructively.

Plot(LTtot);
pos = 1; // 0 is special (no roof tile)
imagedata=0;

for(ctr=start;ctr<=finish;ctr++)
	{
	line = script.line[ctr];
	Rptr=strfirst(strrest(line));
	Rptr=strfirst(line);                       // Get first term
	if(Rptr!=NOTHING)                       // If it exists..
		{
		LTlist[pos].fname = hardfirst(line);     // Get filename
		imagedata += Init_LightMap(pos);
		pos++;  // Go onto the next entry in the sprite list
		}
	}

ilog_printf("\n");
}


/*
 *      Restring - Turn all the static strings from compiled script
 *                 into dynamically-allocated strings (that can be lengthened)
 */

void Restring()
{
int ctr;
char buf[MAX_LINE_LEN];

#define RESTR_MACRO(x) {strcpy(buf,x);\
                       strupr(buf);\
                       x=(char *)M_get(MAX_LINE_LEN,1);\
                       strcpy(x,buf);}

#define I_RESTR_MACRO(x) {strcpy(buf,x);\
                       x=(char *)M_get(MAX_LINE_LEN,1);\
                       strcpy(x,buf);}

for(ctr=0;ctr<SPtot;ctr++)
    {
    RESTR_MACRO(SPlist[ctr].name);
    RESTR_MACRO(SPlist[ctr].fname);
    }

for(ctr=0;ctr<SQtot;ctr++)
    RESTR_MACRO(SQlist[ctr].name);

for(ctr=0;ctr<Songs;ctr++)
    {
    RESTR_MACRO(mustab[ctr].name);
    RESTR_MACRO(mustab[ctr].fname);
    }

for(ctr=0;ctr<CHtot;ctr++)
    {
    RESTR_MACRO(CHlist[ctr].name);
    I_RESTR_MACRO(CHlist[ctr].desc);
    I_RESTR_MACRO(CHlist[ctr].shortdesc);
    }

for(ctr=0;ctr<Waves;ctr++)
    {
    RESTR_MACRO(wavtab[ctr].name);
    RESTR_MACRO(wavtab[ctr].fname);
    }

for(ctr=0;ctr<COtot;ctr++)
    {
    RESTR_MACRO(COlist[ctr].name);
    RESTR_MACRO(COlist[ctr].fname);
    }

for(ctr=0;ctr<TItot;ctr++)
    {
    RESTR_MACRO(TIlist[ctr].name);
    I_RESTR_MACRO(TIlist[ctr].desc);
    }

}

/*
 *      GetVrmName - Return a pointer to the VRM's name string
 */

char *GetVrmName(int num)
{
if(num>0 && num <COtot)
    return COlist[num].name;
return NULL;
}

/*
 *      Dump - Crash out and show a fragment of the code
 */

void Dump(long lineno,char *error,char *help)
{
char errbuf[1024];
char errstr[1024];
short ctr,sl,c2,c2m8;
long l2,Lctr;

//KillGFX();

//ilog_printf("Error: %s\n%s [%d]\n",error,help,lineno);

ilog_printf("\n\n");
irecon_colour(0,200,0); // Green
sprintf(errstr,"Error in %s at line %ld: %s\r\n",compilename,lineno,error);

ilog_printf(errstr);

l2=lineno+5;
if(l2>script.lines)
	l2=script.lines;
irecon_colour(200,0,0); // Red
for(Lctr=lineno;Lctr<l2;Lctr++)
	if(script.line[Lctr]!=NOTHING&&script.line[Lctr]!=NULL)
	{
	c2=0;
	sl=strlen(script.line[Lctr]);
	for(ctr=0;ctr<sl;ctr++)
		if(script.line[Lctr][ctr]=='\t')
			{
			c2m8=c2%8;
			if(!c2m8)
				c2m8=8;
			for(;c2m8>0;c2m8--)
				errbuf[c2++]=' ';
			}
		else
			errbuf[c2++]=script.line[Lctr][ctr];
	errbuf[c2++]=0;
	sprintf(errstr,"%6ld: %s\r\n",Lctr,errbuf);

	ilog_printf(errstr);
	}
irecon_colour(0,200,0); // Green
if(help)
        {
	sprintf(errstr,"\r\n%s\r\n",help);
        ilog_printf(errstr);
        }
ilog_printf("\n");
irecon_colour(200,200,200); // Light gray
ilog_printf("Press a key to exit\n");
WaitForAscii();
//KillGFX();
exit(1);
}

/*
 *      Purify - Remove comments from the script to ease compilation
 */

void Purify()
{
int ctr,len,pos;
char *ptr;

// Scan each line for comments

for(ctr=0;ctr<script.lines;ctr++)
    {
    ptr = strchr(script.line[ctr],'#');         // Look for the hash
    if(ptr)
           *ptr=0;                              // make EOL if we found one

    ptr = strchr(script.line[ctr],10);          // Look for the LF
    if(ptr)
           *ptr=0;                              // make EOL if we found one
    }

// Scan each line for whitespace and turn it into space

for(ctr=0;ctr<script.lines;ctr++)
    {
//    strupr(script.line[ctr]);
    len = strlen(script.line[ctr]);    // Got length of line
    for(pos=0;pos<len;pos++)
        if(isxspace(script.line[ctr][pos]))
                script.line[ctr][pos] = ' ';    // Whitespace becomes space

    strstrip(script.line[ctr]);
    }


return;
}


/*
 *      SeekSect - Locate each discrete section
 */

void SeekSect()
{
int ctr;
char buf[MAX_LINE_LEN];

no_of_sections = 0;

for(ctr=0;ctr<script.lines;ctr++)
	{
//	ilog_quiet("line %d: %x\n",ctr,script.line[ctr]);
//	ilog_quiet("line %d='%s'\n",ctr,script.line[ctr]);
//	strupr(script.line[ctr]);               // Convert to uppercase
	strcpy(Rbuffer,script.line[ctr]);       // Get line
	strcpy(buf,strfirst(Rbuffer));          // Get first word from line
	if(buf[0])
		if(!istricmp(buf,"SECTION:"))
			{
			sect_map[no_of_sections++].end = ctr-1;
			sect_map[no_of_sections].start = ctr;
//			strcpy(sect_map[no_of_sections].name,strfirst(strrest(Rbuffer)));
			strcpy(sect_map[no_of_sections].name,strgetword(Rbuffer,2));
			}
	}
sect_map[no_of_sections++].end = script.lines-1;
}

